from unittest.mock import patch

import pytest
from blueapi.core import BlueskyContext
from dodal.devices.baton import Baton
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.utils.context import find_device_in_context
from mx_bluesky.hyperion.__main__ import main
from mx_bluesky.hyperion.baton_handler import HYPERION_USER


@pytest.fixture(autouse=True)
def patch_setup_devices():
    from mx_bluesky.hyperion.utils.context import setup_devices

    def patched_setup_devices(context: BlueskyContext, dev_mode: bool):
        setup_devices(context, True)
        # reapply requested user to the newly created fake baton
        baton_with_requested_user(context, HYPERION_USER)

    # Patch setup_devices to patch the baton again when it is re-created
    with (
        patch(
            "mx_bluesky.hyperion.baton_handler.setup_devices",
            side_effect=patched_setup_devices,
        ) as patched_func,
        patch(
            "mx_bluesky.hyperion.utils.context.setup_devices",
            side_effect=patched_setup_devices,
        ),
    ):
        yield patched_func


@pytest.fixture(autouse=True)
def patch_udc_default_state():
    with patch("mx_bluesky.hyperion.baton_handler.move_to_udc_default_state"):
        yield


def baton_with_requested_user(
    bluesky_context: BlueskyContext, user: str = HYPERION_USER
) -> Baton:
    baton = find_device_in_context(bluesky_context, "baton", Baton)
    set_mock_value(baton.requested_user, user)
    return baton


@pytest.mark.requires(external="fake_agamemnon")
@patch.dict(
    "os.environ", {"AGAMEMNON_URL": "http://localhost:8088/", "BEAMLINE": "i03"}
)
@patch("sys.argv", new=["hyperion", "--dev", "--mode", "udc"])
@pytest.mark.timeout(0)
def test_udc_mode():
    """
    Run Hyperion in UDC mode against a fake agamemnon server, to test the API.
    One is provided in the hyperion-system-test project, that just continually serves a wait command."""
    main()



from bluesky import plan_stubs as bps
from dodal.devices.detector import (
    DetectorParams,
)
from dodal.devices.i03.dcm import DCM


def fill_in_energy_if_not_supplied(dcm: DCM, detector_params: DetectorParams):
    if not detector_params.expected_energy_ev:
        actual_energy_ev = 1000 * (yield from bps.rd(dcm.energy_in_kev))
        detector_params.expected_energy_ev = actual_energy_ev
    return detector_params



import dataclasses

import pytest
from dodal.devices.zebra.zebra import (
    I03Axes,
    Zebra,
)
from dodal.devices.zebra.zebra_controlled_shutter import (
    ZebraShutter,
    ZebraShutterControl,
)

from mx_bluesky.hyperion.device_setup_plans.setup_zebra import (
    setup_zebra_for_panda_flyscan,
    setup_zebra_for_rotation,
)
from mx_bluesky.phase1_zebra.device_setup_plans.setup_zebra import (
    configure_zebra_and_shutter_for_auto_shutter,
    setup_zebra_for_gridscan,
    tidy_up_zebra_after_gridscan,
)


async def _get_shutter_input_2(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[2]
        .get_value()
    )


async def _get_shutter_input_1(zebra: Zebra):
    return (
        await zebra.logic_gates.and_gates[zebra.mapping.AND_GATE_FOR_AUTO_SHUTTER]
        .sources[1]
        .get_value()
    )


async def test_zebra_set_up_for_panda_gridscan(
    RE, zebra: Zebra, zebra_shutter: ZebraShutter
):
    RE(setup_zebra_for_panda_flyscan(zebra, zebra_shutter, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.IN1_TTL
    )
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_PANDA].get_value()
        == zebra.mapping.sources.IN3_TTL
    )
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1


async def test_zebra_set_up_for_gridscan(RE, zebra: Zebra, zebra_shutter: ZebraShutter):
    @dataclasses.dataclass
    class Composite:
        zebra: Zebra
        sample_shutter: ZebraShutter

    composite = Composite(zebra, zebra_shutter)
    RE(setup_zebra_for_gridscan(composite, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.IN3_TTL
    )
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1


async def test_zebra_set_up_for_rotation(RE, zebra: Zebra, zebra_shutter: ZebraShutter):
    RE(setup_zebra_for_rotation(zebra, zebra_shutter, wait=True))
    assert await zebra.pc.gate_trigger.get_value() == I03Axes.OMEGA.value
    assert await zebra.pc.gate_width.get_value() == pytest.approx(360, 0.01)
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1


async def test_zebra_cleanup(RE, zebra: Zebra, zebra_shutter: ZebraShutter):
    RE(tidy_up_zebra_after_gridscan(zebra, zebra_shutter, wait=True))
    assert (
        await zebra.output.out_pvs[zebra.mapping.outputs.TTL_DETECTOR].get_value()
        == zebra.mapping.sources.PC_PULSE
    )
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.PC_GATE


class MyException(Exception):
    pass


async def test_configure_zebra_and_shutter_for_auto(
    RE, zebra: Zebra, zebra_shutter: ZebraShutter
):
    RE(
        configure_zebra_and_shutter_for_auto_shutter(
            zebra, zebra_shutter, zebra.mapping.sources.IN4_TTL
        )
    )
    assert await zebra_shutter.control_mode.get_value() == ZebraShutterControl.AUTO
    assert await _get_shutter_input_1(zebra) == zebra.mapping.sources.SOFT_IN1
    assert await _get_shutter_input_2(zebra) == zebra.mapping.sources.IN4_TTL



"""
A minimal test beamline that contains only a baton, for use in tests which need a beamline
but not all the devices, so that test execution isn't slowed down by loading lots of
python modules/creating objects.
See Also:
    use_beamline_t01()
"""

from dodal.common.beamlines.beamline_utils import device_factory
from dodal.devices.baton import Baton
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.utils import BeamlinePrefix, get_beamline_name

BL = get_beamline_name("t01")
PREFIX = BeamlinePrefix(BL)


@device_factory()
def baton() -> Baton:
    """Get the i03 baton device, instantiate it if it hasn't already been.
    If this is called when already instantiated in i03, it will return the existing object.
    """
    return Baton(f"{PREFIX.beamline_prefix}-CS-BATON-01:")


@device_factory()
def xbpm_feedback() -> XBPMFeedback:
    """Get the i03 XBPM feeback device, instantiate it if it hasn't already been.
    If this is called when already instantiated in i03, it will return the existing object.
    """
    return XBPMFeedback(
        PREFIX.beamline_prefix,
        "xbpm_feedback",
    )



from dodal.utils import get_beamline_name
from event_model import Event, EventDescriptor, RunStart

from mx_bluesky.common.external_interaction.alerting import (
    Metadata,
    get_alerting_service,
)
from mx_bluesky.common.external_interaction.callbacks.common.plan_reactive_callback import (
    PlanReactiveCallback,
)
from mx_bluesky.common.utils.log import ISPYB_ZOCALO_CALLBACK_LOGGER
from mx_bluesky.hyperion.parameters.constants import CONST


class AlertOnContainerChange(PlanReactiveCallback):
    """Sends an alert to beamline staff when a pin from a new puck has been loaded.
    This tends to be used as a heartbeat so we know that UDC is running."""

    def __init__(self):
        super().__init__(log=ISPYB_ZOCALO_CALLBACK_LOGGER)
        self._new_container = None
        self._visit = None
        self._sample_id = None
        self.descriptors: dict[str, EventDescriptor] = {}

    def activity_gated_descriptor(self, doc: EventDescriptor) -> EventDescriptor | None:
        self.descriptors[doc["uid"]] = doc
        return super().activity_gated_descriptor(doc)

    def activity_gated_event(self, doc: Event) -> Event | None:
        event_descriptor = self.descriptors.get(doc["descriptor"])
        if (
            event_descriptor
            and event_descriptor.get("name") == CONST.DESCRIPTORS.ROBOT_PRE_LOAD
        ):
            current_container = int(doc["data"]["robot-current_puck"])
            if self._new_container != current_container:
                beamline = get_beamline_name("")
                get_alerting_service().raise_alert(
                    f"UDC moved on to puck {self._new_container} on {beamline}",
                    f"Hyperion finished container {current_container} and moved on to {self._new_container}",
                    {
                        Metadata.SAMPLE_ID: str(self._sample_id),
                        Metadata.VISIT: self._visit or "",
                        Metadata.CONTAINER: str(self._new_container),
                    },
                )
        return doc

    def activity_gated_start(self, doc: RunStart):
        metadata = doc.get("metadata", {})
        self._new_container = metadata.get("container")
        self._sample_id = metadata.get("sample_id")
        self._visit = metadata.get("visit")



import json
from unittest.mock import MagicMock, call, patch

import pytest
from pydantic.dataclasses import dataclass

from mx_bluesky.common.external_interaction.config_server import MXConfigClient
from mx_bluesky.common.parameters.constants import (
    GDA_DOMAIN_PROPERTIES_PATH,
    FeatureSetting,
    FeatureSettingources,
    OavConstants,
)
from mx_bluesky.hyperion.external_interaction.config_server import (
    get_hyperion_config_client,
)
from mx_bluesky.hyperion.parameters.constants import HyperionFeatureSetting


def test_verify_feature_parameters():
    class BadHyperionFeatureSettingources(FeatureSettingources):
        USE_GPU_RESULTS = "gda.mx.hyperion.xrc.use_gpu_results"
        USE_ZEBRA_FOR_GRIDSCAN = "gda.mx.hyperion.use_panda_for_gridscans"
        SET_STUB_OFFSETS = "gda.mx.hyperion.do_stub_offsets"

    with pytest.raises(AssertionError):
        MXConfigClient(
            feature_sources=BadHyperionFeatureSettingources,
            feature_dc=HyperionFeatureSetting,
        )


@patch("mx_bluesky.common.external_interaction.config_server.LOGGER.warning")
def test_get_json_config_good_request(mock_log_warn: MagicMock):
    with open(OavConstants.OAV_CONFIG_JSON) as f:
        expected_dict = json.loads(f.read())
    assert expected_dict == get_hyperion_config_client().get_json_config(
        OavConstants.OAV_CONFIG_JSON
    )
    mock_log_warn.assert_not_called()


@patch("mx_bluesky.common.external_interaction.config_server.LOGGER.warning")
def test_get_json_config_on_bad_request(mock_log_warn: MagicMock):
    with open(OavConstants.OAV_CONFIG_JSON) as f:
        expected_dict = json.loads(f.read())
    server = get_hyperion_config_client()
    server.get_file_contents = MagicMock(side_effect=Exception)
    assert expected_dict == server.get_json_config(OavConstants.OAV_CONFIG_JSON)
    mock_log_warn.assert_called_once()


@patch(
    "mx_bluesky.common.external_interaction.config_server.GDA_DOMAIN_PROPERTIES_PATH",
    new="tests/test_data/test_domain_properties_with_no_gpu",
)
@patch("mx_bluesky.common.external_interaction.config_server.LOGGER.warning")
def test_get_feature_flags_good_request(mock_log_warn: MagicMock):
    expected_features_dict = {
        "USE_GPU_RESULTS": False,
        "USE_PANDA_FOR_GRIDSCAN": True,
        "SET_STUB_OFFSETS": False,
    }
    server = get_hyperion_config_client()
    assert server.get_feature_flags() == HyperionFeatureSetting(
        **expected_features_dict
    )
    mock_log_warn.assert_not_called()


def test_get_feature_flags_cache():
    server = get_hyperion_config_client()
    expected_features_dict = {
        "USE_GPU_RESULTS": False,
        "USE_PANDA_FOR_GRIDSCAN": True,
        "SET_STUB_OFFSETS": False,
    }
    expected_features = HyperionFeatureSetting(**expected_features_dict)
    server._cached_features = expected_features
    with patch(
        "mx_bluesky.common.external_interaction.config_server.MXConfigClient.get_file_contents"
    ) as mock_get_file_contents:
        assert server.get_feature_flags() == expected_features
        mock_get_file_contents.assert_not_called()
        server._get_feature_flags(reset_cached_result=True)
        mock_get_file_contents.assert_called_once()


def test_get_json_config_cache():
    server = get_hyperion_config_client()
    with open(OavConstants.OAV_CONFIG_JSON) as f:
        expected_dict = json.loads(f.read())
    get_hyperion_config_client().get_json_config(OavConstants.OAV_CONFIG_JSON)
    assert server._cached_json_config[OavConstants.OAV_CONFIG_JSON] == expected_dict
    with patch(
        "mx_bluesky.common.external_interaction.config_server.MXConfigClient.get_file_contents"
    ) as mock_get_file_contents:
        server.get_json_config(OavConstants.OAV_CONFIG_JSON)
        mock_get_file_contents.assert_not_called()
        server._get_json_config(OavConstants.OAV_CONFIG_JSON, reset_cached_result=True)
        mock_get_file_contents.assert_called_once()


@patch(
    "mx_bluesky.common.external_interaction.config_server.GDA_DOMAIN_PROPERTIES_PATH",
    new="BAD_PATH",
)
@patch("mx_bluesky.common.external_interaction.config_server.LOGGER.warning")
def test_get_feature_flags_bad_request(mock_log_warn: MagicMock):
    server = get_hyperion_config_client()
    assert server.get_feature_flags() == HyperionFeatureSetting()
    mock_log_warn.assert_called_once()


def test_refresh_cache():
    server = get_hyperion_config_client()
    server.get_file_contents = MagicMock()
    server.refresh_cache()
    call_list = [
        call(GDA_DOMAIN_PROPERTIES_PATH, reset_cached_result=True),
        call(OavConstants.OAV_CONFIG_JSON, dict, reset_cached_result=True),
    ]
    server.get_file_contents.assert_has_calls(call_list, any_order=True)


class BadFeatureSettingSources(FeatureSettingources):
    USE_GPU_RESULTS = "gda.mx.hyperion.xrc.use_gpu_results"
    USE_PANDA_FOR_GRIDSCAN = "gda.mx.hyperion.use_panda_for_gridscans"
    SET_STUB_OFFSETS = "gda.mx.hyperion.do_stub_offsets"
    PANDA_RUNUP_DISTANCE_MM = "gda.mx.hyperion.panda_runup_distance_mm"
    MISSING_FEATURE = "gda.mx.hyperion.missing_feature"


@dataclass
class BadFeatureSetting(FeatureSetting):
    USE_GPU_RESULTS: bool = True
    USE_PANDA_FOR_GRIDSCAN: bool = False
    SET_STUB_OFFSETS: bool = False
    MISSING_FEATURE: bool = False
    PANDA_RUNUP_DISTANCE_MM: float = 0.16


@patch("mx_bluesky.common.external_interaction.config_server.LOGGER.warning")
def test_warning_on_missing_features_in_file(mock_log_warn: MagicMock):
    server = MXConfigClient(BadFeatureSettingSources, BadFeatureSetting)

    expected_features_dict = {
        "USE_GPU_RESULTS": True,
        "USE_PANDA_FOR_GRIDSCAN": False,
        "SET_STUB_OFFSETS": False,
        "PANDA_RUNUP_DISTANCE_MM": 0.16,
        "MISSING_FEATURE": False,
    }
    assert server.get_feature_flags() == BadFeatureSetting(**expected_features_dict)
    assert (
        "MISSING_FEATURE" in mock_log_warn.call_args_list[0][0][0]
    )  # call -> tuple -> contents
    mock_log_warn.assert_called_once()



from __future__ import annotations

from functools import partial

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.common_dcm import BaseDCM
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import (
    ZebraFastGridScan,
    set_fast_grid_scan_params,
)
from dodal.devices.flux import Flux
from dodal.devices.mx_phase1.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import Undulator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from dodal.plans.preprocessors.verify_undulator_gap import (
    verify_undulator_gap_before_run_decorator,
)

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    construct_beamline_specific_FGS_features,
)
from mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan import (
    grid_detect_then_xray_centre,
)
from mx_bluesky.common.experiment_plans.oav_snapshot_plan import (
    setup_beamline_for_OAV,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.parameters.constants import (
    EnvironmentConstants,
    OavConstants,
    PlanGroupCheckpointConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.device_composites import (
    GridDetectThenXRayCentreComposite,
)
from mx_bluesky.common.parameters.gridscan import GridCommon, SpecifiedThreeDGridScan
from mx_bluesky.common.preprocessors.preprocessors import (
    transmission_and_xbpm_feedback_for_collection_decorator,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.phase1_zebra.device_setup_plans.setup_zebra import (
    setup_zebra_for_gridscan,
    tidy_up_zebra_after_gridscan,
)


# See https://github.com/DiamondLightSource/blueapi/issues/506 for using device composites
def i04_grid_detect_then_xray_centre(
    parameters: GridCommon,
    aperture_scatterguard: ApertureScatterguard = inject("aperture_scatterguard"),
    attenuator: BinaryFilterAttenuator = inject("attenuator"),
    backlight: Backlight = inject("backlight"),
    beamstop: Beamstop = inject("beamstop"),
    dcm: BaseDCM = inject("dcm"),
    zebra_fast_grid_scan: ZebraFastGridScan = inject("zebra_fast_grid_scan"),
    flux: Flux = inject("flux"),
    oav: OAV = inject("oav"),
    pin_tip_detection: PinTipDetection = inject("pin_tip_detection"),
    s4_slit_gaps: S4SlitGaps = inject("s4_slit_gaps"),
    undulator: Undulator = inject("undulator"),
    xbpm_feedback: XBPMFeedback = inject("xbpm_feedback"),
    zebra: Zebra = inject("zebra"),
    robot: BartRobot = inject("robot"),
    sample_shutter: ZebraShutter = inject("sample_shutter"),
    eiger: EigerDetector = inject("eiger"),
    synchrotron: Synchrotron = inject("synchrotron"),
    zocalo: ZocaloResults = inject("zocalo"),
    smargon: Smargon = inject("smargon"),
    detector_motion: DetectorMotion = inject("detector_motion"),
    oav_config: str = OavConstants.OAV_CONFIG_JSON,
    udc: bool = False,
) -> MsgGenerator:
    """
    A composite plan which:
    - Uses the OAV to draw a virtual grid over the sample and to take snapshots of the sample
    - Scans through the grid to identify the crystal centre
    - Changes the aperture to match the beam size to the crystal size
    - Moves the sample to the crystal centre of mass


    i04's implementation of this plan is very similar to Hyperion. However, since i04
    isn't running in a continious Bluesky UDC loop, we take additional steps in beamline
    tidy-up.
    """

    composite = GridDetectThenXRayCentreComposite(
        eiger,
        synchrotron,
        zocalo,
        smargon,
        aperture_scatterguard,
        attenuator,
        backlight,
        beamstop,
        dcm,
        detector_motion,
        zebra_fast_grid_scan,
        flux,
        oav,
        pin_tip_detection,
        s4_slit_gaps,
        undulator,
        xbpm_feedback,
        zebra,
        robot,
        sample_shutter,
    )

    def tidy_beamline_if_not_udc():
        if not udc:
            yield from get_ready_for_oav_and_close_shutter(
                composite.smargon,
                composite.backlight,
                composite.aperture_scatterguard,
                composite.detector_motion,
            )

    @bpp.finalize_decorator(tidy_beamline_if_not_udc)
    def _inner_grid_detect_then_xrc():
        # These callbacks let us talk to ISPyB and Nexgen. They aren't included in the common plan because
        # Hyperion handles its callbacks differently to BlueAPI-managed plans, see
        # https://github.com/DiamondLightSource/mx-bluesky/issues/1117
        callbacks = create_gridscan_callbacks()

        @bpp.subs_decorator(callbacks)
        @verify_undulator_gap_before_run_decorator(composite)
        @transmission_and_xbpm_feedback_for_collection_decorator(
            composite, parameters.transmission_frac, PlanNameConstants.GRIDSCAN_OUTER
        )
        def grid_detect_then_xray_centre_with_callbacks():
            yield from grid_detect_then_xray_centre(
                composite=composite,
                parameters=parameters,
                xrc_params_type=SpecifiedThreeDGridScan,
                construct_beamline_specific=construct_i04_specific_features,
                oav_config=oav_config,
            )

        yield from grid_detect_then_xray_centre_with_callbacks()

    yield from _inner_grid_detect_then_xrc()


def get_ready_for_oav_and_close_shutter(
    smargon: Smargon,
    backlight: Backlight,
    aperture_scatterguard: ApertureScatterguard,
    detector_motion: DetectorMotion,
):
    yield from bps.wait(PlanGroupCheckpointConstants.GRID_READY_FOR_DC)
    group = "get_ready_for_oav_and_close_shutter"
    LOGGER.info("Non-udc tidy: Seting up beamline for OAV")
    yield from setup_beamline_for_OAV(
        smargon, backlight, aperture_scatterguard, group=group
    )
    LOGGER.info("Non-udc tidy: Closing detector shutter")
    yield from bps.abs_set(
        detector_motion.shutter,
        0,
        group=group,
    )
    yield from bps.wait(group)


def create_gridscan_callbacks() -> tuple[
    GridscanNexusFileCallback, GridscanISPyBCallback
]:
    return (
        GridscanNexusFileCallback(param_type=SpecifiedThreeDGridScan),
        GridscanISPyBCallback(
            param_type=GridCommon,
            emit=ZocaloCallback(
                PlanNameConstants.DO_FGS, EnvironmentConstants.ZOCALO_ENV
            ),
        ),
    )


def construct_i04_specific_features(
    xrc_composite: GridDetectThenXRayCentreComposite,
    xrc_parameters: SpecifiedThreeDGridScan,
) -> BeamlineSpecificFGSFeatures:
    """
    Get all the information needed to do the i04 XRC flyscan.
    """
    signals_to_read_pre_flyscan = [
        xrc_composite.undulator.current_gap,
        xrc_composite.synchrotron.synchrotron_mode,
        xrc_composite.s4_slit_gaps.xgap,
        xrc_composite.s4_slit_gaps.ygap,
        xrc_composite.smargon.x,
        xrc_composite.smargon.y,
        xrc_composite.smargon.z,
        xrc_composite.dcm.energy_in_kev,
    ]

    signals_to_read_during_collection = [
        xrc_composite.aperture_scatterguard,
        xrc_composite.attenuator.actual_transmission,
        xrc_composite.flux.flux_reading,
        xrc_composite.dcm.energy_in_kev,
        xrc_composite.eiger.bit_depth,
    ]

    tidy_plan = partial(
        tidy_up_zebra_after_gridscan,
        xrc_composite.zebra,
        xrc_composite.sample_shutter,
        group="flyscan_zebra_tidy",
        wait=True,
    )
    set_flyscan_params_plan = partial(
        set_fast_grid_scan_params,
        xrc_composite.zebra_fast_grid_scan,
        xrc_parameters.FGS_params,
    )
    fgs_motors = xrc_composite.zebra_fast_grid_scan
    return construct_beamline_specific_FGS_features(
        setup_zebra_for_gridscan,
        tidy_plan,
        set_flyscan_params_plan,
        fgs_motors,
        signals_to_read_pre_flyscan,
        signals_to_read_during_collection,
        get_xrc_results_from_zocalo=True,
    )



from __future__ import annotations

import dataclasses
from collections.abc import Callable, Sequence
from functools import partial

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
import numpy as np
from bluesky.protocols import Readable
from bluesky.utils import MsgGenerator
from dodal.common.beamlines.commissioning_mode import read_commissioning_mode
from dodal.devices.fast_grid_scan import (
    FastGridScanCommon,
)
from dodal.devices.zocalo import ZocaloResults
from dodal.devices.zocalo.zocalo_results import (
    XrcResult,
    get_full_processing_results,
)

from mx_bluesky.common.experiment_plans.inner_plans.do_fgs import (
    ZOCALO_STAGE_GROUP,
    kickoff_and_complete_gridscan,
)
from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_plan,
)
from mx_bluesky.common.parameters.constants import (
    DocDescriptorNames,
    GridscanParamConstants,
    PlanGroupCheckpointConstants,
    PlanNameConstants,
)
from mx_bluesky.common.parameters.device_composites import FlyScanEssentialDevices
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.utils.exceptions import (
    CrystalNotFoundException,
    SampleException,
)
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.utils.tracing import TRACER
from mx_bluesky.common.xrc_result import XRayCentreResult


@dataclasses.dataclass
class BeamlineSpecificFGSFeatures:
    setup_trigger_plan: Callable[..., MsgGenerator]
    tidy_plan: Callable[..., MsgGenerator]
    set_flyscan_params_plan: Callable[..., MsgGenerator]
    fgs_motors: FastGridScanCommon
    read_pre_flyscan_plan: Callable[
        ..., MsgGenerator
    ]  # Eventually replace with https://github.com/DiamondLightSource/mx-bluesky/issues/819
    read_during_collection_plan: Callable[..., MsgGenerator]
    get_xrc_results_from_zocalo: bool


def generic_tidy(xrc_composite: FlyScanEssentialDevices, wait=True) -> MsgGenerator:
    """Tidy Zocalo and turn off Eiger dev/shm. Ran after the beamline-specific tidy plan"""

    LOGGER.info("Tidying up Zocalo")
    group = "generic_tidy"
    # make sure we don't consume any other results
    yield from bps.unstage(xrc_composite.zocalo, group=group)

    # Turn off dev/shm streaming to avoid filling disk, see https://github.com/DiamondLightSource/hyperion/issues/1395
    LOGGER.info("Turning off Eiger dev/shm streaming")
    # Fix types in ophyd-async (https://github.com/DiamondLightSource/mx-bluesky/issues/855)
    yield from bps.abs_set(
        xrc_composite.eiger.odin.fan.dev_shm_enable,  # type: ignore
        0,
        group=group,
    )
    yield from bps.wait(group)


def construct_beamline_specific_FGS_features(
    setup_trigger_plan: Callable[..., MsgGenerator],
    tidy_plan: Callable[..., MsgGenerator],
    set_flyscan_params_plan: Callable[..., MsgGenerator],
    fgs_motors: FastGridScanCommon,
    signals_to_read_pre_flyscan: list[Readable],
    signals_to_read_during_collection: list[Readable],
    get_xrc_results_from_zocalo: bool = False,
) -> BeamlineSpecificFGSFeatures:
    """Construct the class needed to do beamline-specific parts of the XRC FGS

    Args:
        setup_trigger_plan (Callable): Configure triggering, for example with the Zebra or PandA device.
        Ran directly before kicking off the gridscan.

        tidy_plan (Callable): Tidy up states of devices. Ran at the end of the flyscan, regardless of
        whether or not it finished successfully. Zocalo and Eiger are cleaned up separately

        set_flyscan_params_plan (Callable): Set PV's for the relevant Fast Grid Scan dodal device

        fgs_motors (Callable): Composite device representing the fast grid scan's motion program parameters.

        signals_to_read_pre_flyscan (Callable): Signals which will be read and saved as a bluesky event document
        after all configuration, but before the gridscan.

        signals_to_read_during_collection (Callable): Signals which will be read and saved as a bluesky event
        document whilst the gridscan motion is in progress

        get_xrc_results_from_zocalo (bool): If true, fetch grid scan results from zocalo after completion, as well as
        update the ispyb comment field with information about the results. See _fetch_xrc_results_from_zocalo
    """
    read_pre_flyscan_plan = partial(
        read_hardware_plan,
        signals_to_read_pre_flyscan,
        DocDescriptorNames.HARDWARE_READ_PRE,
    )

    read_during_collection_plan = partial(
        read_hardware_plan,
        signals_to_read_during_collection,
        DocDescriptorNames.HARDWARE_READ_DURING,
    )

    return BeamlineSpecificFGSFeatures(
        setup_trigger_plan,
        tidy_plan,
        set_flyscan_params_plan,
        fgs_motors,
        read_pre_flyscan_plan,
        read_during_collection_plan,
        get_xrc_results_from_zocalo,
    )


def common_flyscan_xray_centre(
    composite: FlyScanEssentialDevices,
    parameters: SpecifiedThreeDGridScan,
    beamline_specific: BeamlineSpecificFGSFeatures,
) -> MsgGenerator:
    """Main entry point of the MX-Bluesky x-ray centering flyscan

    Args:
        composite (FlyScanEssentialDevices): Devices required to perform this plan.

        parameters (SpecifiedThreeDGridScan): Parameters required to perform this plan.

        beamline_specific (BeamlineSpecificFGSFeatures): Configure the beamline-specific version
        of this plan: For example triggering setup and tidy up plans, as well as what to do with the
        centering results.

    With a minimum set of devices and parameters, prepares for; performs; and tidies up a flyscan
    x-ray-center plan. This includes: Configuring desired triggering; writing nexus files; triggering zocalo;
    reading hardware before and during the scan; and tidying up devices after
    the plan is complete. Optionally fetch results from zocalo after completing the grid scan.

    This plan will also push data to ispyb when used with the ispyb_activation_decorator.

    There are a few other useful decorators to use with this plan, see: verify_undulator_gap_before_run_decorator, transmission_and_xbpm_feedback_for_collection_decorator
    """

    def _overall_tidy():
        yield from beamline_specific.tidy_plan()
        yield from generic_tidy(composite)

    def _decorated_flyscan():
        @bpp.set_run_key_decorator(PlanNameConstants.GRIDSCAN_OUTER)
        @bpp.run_decorator(  # attach experiment metadata to the start document
            md={
                "subplan_name": PlanNameConstants.GRIDSCAN_OUTER,
                "mx_bluesky_parameters": parameters.model_dump_json(),
                "activate_callbacks": [
                    "GridscanNexusFileCallback",
                ],
            }
        )
        @bpp.finalize_decorator(lambda: _overall_tidy())
        def run_gridscan_and_tidy(
            fgs_composite: FlyScanEssentialDevices,
            params: SpecifiedThreeDGridScan,
            beamline_specific: BeamlineSpecificFGSFeatures,
        ) -> MsgGenerator:
            yield from beamline_specific.setup_trigger_plan(fgs_composite, parameters)

            LOGGER.info("Starting grid scan")
            yield from bps.stage(
                fgs_composite.zocalo, group=ZOCALO_STAGE_GROUP
            )  # connect to zocalo and make sure the queue is clear
            yield from run_gridscan(fgs_composite, params, beamline_specific)

            LOGGER.info("Grid scan finished")

            if beamline_specific.get_xrc_results_from_zocalo:
                yield from _fetch_xrc_results_from_zocalo(composite.zocalo, parameters)

        yield from run_gridscan_and_tidy(composite, parameters, beamline_specific)

    composite.eiger.set_detector_parameters(parameters.detector_params)
    yield from _decorated_flyscan()


def _fetch_xrc_results_from_zocalo(
    zocalo_results: ZocaloResults,
    parameters: SpecifiedThreeDGridScan,
) -> MsgGenerator:
    """
    Get XRC results from the ZocaloResults device which was staged during a grid scan,
    and store them in XRayCentreEventHandler.xray_centre_results by firing an event.

    The RunEngine must be subscribed to XRayCentreEventHandler for this plan to work.
    """

    LOGGER.info("Getting X-ray center Zocalo results...")

    yield from bps.trigger(zocalo_results)
    LOGGER.info("Zocalo triggered and read, interpreting results.")
    xrc_results = yield from get_full_processing_results(zocalo_results)
    LOGGER.info(f"Got xray centres, top 5: {xrc_results[:5]}")
    filtered_results = [
        result
        for result in xrc_results
        if result["total_count"]
        >= GridscanParamConstants.ZOCALO_MIN_TOTAL_COUNT_THRESHOLD
    ]
    discarded_count = len(xrc_results) - len(filtered_results)
    if discarded_count > 0:
        LOGGER.info(f"Removed {discarded_count} results because below threshold")
    if filtered_results:
        flyscan_results = [
            _xrc_result_in_boxes_to_result_in_mm(xr, parameters)
            for xr in filtered_results
        ]
    else:
        commissioning_mode = yield from read_commissioning_mode()
        if commissioning_mode:
            LOGGER.info("Commissioning mode enabled, returning dummy result")
            flyscan_results = [_generate_dummy_xrc_result(parameters)]
        else:
            LOGGER.warning("No X-ray centre received")
            raise CrystalNotFoundException()
    yield from _fire_xray_centre_result_event(flyscan_results)


def _generate_dummy_xrc_result(params: SpecifiedThreeDGridScan) -> XRayCentreResult:
    com = [params.x_steps / 2, params.y_steps / 2, params.z_steps / 2]
    max_voxel = [round(p) for p in com]
    return _xrc_result_in_boxes_to_result_in_mm(
        XrcResult(
            centre_of_mass=com,
            max_voxel=max_voxel,
            bounding_box=[max_voxel, [p + 1 for p in max_voxel]],
            n_voxels=1,
            max_count=10000,
            total_count=100000,
            sample_id=params.sample_id,
        ),
        params,
    )


@bpp.set_run_key_decorator(PlanNameConstants.GRIDSCAN_MAIN)
@bpp.run_decorator(md={"subplan_name": PlanNameConstants.GRIDSCAN_MAIN})
def run_gridscan(
    fgs_composite: FlyScanEssentialDevices,
    parameters: SpecifiedThreeDGridScan,
    beamline_specific: BeamlineSpecificFGSFeatures,
):
    # Currently gridscan only works for omega 0, see https://github.com/DiamondLightSource/mx-bluesky/issues/410
    with TRACER.start_span("moving_omega_to_0"):
        yield from bps.abs_set(fgs_composite.smargon.omega, 0)

    with TRACER.start_span("ispyb_hardware_readings"):
        yield from beamline_specific.read_pre_flyscan_plan()

    LOGGER.info("Setting fgs params")
    yield from beamline_specific.set_flyscan_params_plan()

    LOGGER.info("Waiting for gridscan validity check")
    yield from wait_for_gridscan_valid(beamline_specific.fgs_motors)

    LOGGER.info("Waiting for arming to finish")
    yield from bps.wait(PlanGroupCheckpointConstants.GRID_READY_FOR_DC)
    yield from bps.stage(fgs_composite.eiger)  # type: ignore # See: https://github.com/bluesky/bluesky/issues/1809

    yield from kickoff_and_complete_gridscan(
        beamline_specific.fgs_motors,
        fgs_composite.eiger,
        fgs_composite.synchrotron,
        [parameters.scan_points_first_grid, parameters.scan_points_second_grid],
        parameters.scan_indices,
        plan_during_collection=beamline_specific.read_during_collection_plan,
    )

    # GDA's gridscans requires Z steps to be at 0, so make sure we leave this device
    # in a GDA-happy state.
    yield from bps.abs_set(beamline_specific.fgs_motors.z_steps, 0, wait=False)


def wait_for_gridscan_valid(fgs_motors: FastGridScanCommon, timeout=0.5):
    LOGGER.info("Waiting for valid fgs_params")
    SLEEP_PER_CHECK = 0.1
    times_to_check = int(timeout / SLEEP_PER_CHECK)
    for _ in range(times_to_check):
        scan_invalid = yield from bps.rd(fgs_motors.scan_invalid)
        pos_counter = yield from bps.rd(fgs_motors.position_counter)
        LOGGER.debug(
            f"Scan invalid: {scan_invalid} and position counter: {pos_counter}"
        )
        if not scan_invalid and pos_counter == 0:
            LOGGER.info("Gridscan scan valid and position counter reset")
            return
        yield from bps.sleep(SLEEP_PER_CHECK)
    raise SampleException("Scan invalid - pin too long/short/bent and out of range")


def _xrc_result_in_boxes_to_result_in_mm(
    xrc_result: XrcResult, parameters: SpecifiedThreeDGridScan
) -> XRayCentreResult:
    fgs_params = parameters.FGS_params
    xray_centre = fgs_params.grid_position_to_motor_position(
        np.array(xrc_result["centre_of_mass"])
    )
    # A correction is applied to the bounding box to map discrete grid coordinates to
    # the corners of the box in motor-space; we do not apply this correction
    # to the xray-centre as it is already in continuous space and the conversion has
    # been performed already
    # In other words, xrc_result["bounding_box"] contains the position of the box centre,
    # so we subtract half a box to get the corner of the box
    return XRayCentreResult(
        centre_of_mass_mm=xray_centre,
        bounding_box_mm=(
            fgs_params.grid_position_to_motor_position(
                np.array(xrc_result["bounding_box"][0]) - 0.5
            ),
            fgs_params.grid_position_to_motor_position(
                np.array(xrc_result["bounding_box"][1]) - 0.5
            ),
        ),
        max_count=xrc_result["max_count"],
        total_count=xrc_result["total_count"],
        sample_id=xrc_result["sample_id"],
    )


def _fire_xray_centre_result_event(results: Sequence[XRayCentreResult]):
    def empty_plan():
        return iter([])

    yield from bpp.set_run_key_wrapper(
        bpp.run_wrapper(
            empty_plan(),
            md={
                PlanNameConstants.FLYSCAN_RESULTS: [
                    dataclasses.asdict(r) for r in results
                ]
            },
        ),
        PlanNameConstants.FLYSCAN_RESULTS,
    )



import copy
import json
from datetime import timedelta
from typing import TypedDict

from bluesky.callbacks import CallbackBase
from dodal.log import LOGGER
from event_model.documents import Event, RunStart, RunStop
from redis import StrictRedis


class OmegaReading(TypedDict):
    value: float
    timestamp: float


def extrapolate_omega(
    latest_omega: OmegaReading, previous_omega: OmegaReading, image_timestamp: float
) -> float:
    """Extrapolate an image omega from previous omegas.

    There are a number of assumptions in this calculation:
    * The speed of the smargon is fixed
    * The timestamps from the two different devices are synchronised and match the data
      exactly

    These are accepted to be reasonable based on larger errors likely coming from murko
    itself and that the results ultimately will be averaged out.
    """
    omega_per_sec = (latest_omega["value"] - previous_omega["value"]) / (
        latest_omega["timestamp"] - previous_omega["timestamp"]
    )
    time_elapsed = image_timestamp - latest_omega["timestamp"]
    return latest_omega["value"] + time_elapsed * omega_per_sec


class MurkoCallback(CallbackBase):
    """A callback that triggers murko processing of images.

    It combines metadata readings from e.g the goniometer rotation with the uuid's given
    to us by an `OAVToRedisForwarder` (which describe the location of images in redis).
    And writes these as a package to redis. A separate service then forwards this to murko.

    The metadata and image data arrive independently, it is expected that the image data
    is arriving at a faster rate than gonio metadata and so the value of omega for when
    the image arrives is extrapolated based on previous omega readings.
    """

    DATA_EXPIRY_DAYS = 7

    def __init__(self, redis_host: str, redis_password: str, redis_db: int = 0):
        self.redis_client = StrictRedis(
            host=redis_host, password=redis_password, db=redis_db
        )
        self.last_uuid = None
        self.previous_omegas: list[OmegaReading] = []

    def start(self, doc: RunStart) -> RunStart | None:
        self.sample_id = doc.get("sample_id")
        self.murko_metadata = {
            "zoom_percentage": doc.get("zoom_percentage"),
            "microns_per_x_pixel": doc.get("microns_per_x_pixel"),
            "microns_per_y_pixel": doc.get("microns_per_y_pixel"),
            "beam_centre_i": doc.get("beam_centre_i"),
            "beam_centre_j": doc.get("beam_centre_j"),
            "sample_id": self.sample_id,
        }
        self.last_uuid = None
        self.previous_omegas = []
        LOGGER.info(f"Starting to stream metadata to murko under {self.sample_id}")
        return doc

    def event(self, doc: Event) -> Event:
        if latest_omega := doc["data"].get("smargon-omega"):
            if len(self.previous_omegas) <= 2 and self.last_uuid:
                # For the first few images there's not enough data to extrapolate so we
                # match them one to one
                self.call_murko(self.last_uuid, latest_omega)
            self.previous_omegas.append(
                OmegaReading(
                    value=latest_omega,
                    timestamp=doc["timestamps"]["smargon-omega"],
                )
            )
        elif (uuid := doc["data"].get("oav_to_redis_forwarder-uuid")) is not None:
            if len(self.previous_omegas) >= 2:
                omega = extrapolate_omega(
                    self.previous_omegas[-1],
                    self.previous_omegas[-2],
                    doc["timestamps"]["oav_to_redis_forwarder-uuid"],
                )
                LOGGER.info(f"Using extrapolated omega of {omega}")
                self.call_murko(uuid, omega)
            self.last_uuid = uuid
        return doc

    def call_murko(self, uuid: str, omega: float):
        metadata = copy.deepcopy(self.murko_metadata)
        metadata["omega_angle"] = omega
        metadata["uuid"] = uuid

        # Send metadata to REDIS and trigger murko
        redis_key = f"murko:{metadata['sample_id']}:metadata"
        self.redis_client.hset(redis_key, uuid, json.dumps(metadata))
        self.redis_client.expire(redis_key, timedelta(days=self.DATA_EXPIRY_DAYS))
        self.redis_client.publish("murko", json.dumps(metadata))

    def stop(self, doc: RunStop) -> RunStop | None:
        LOGGER.info(f"Finished streaming {self.sample_id} to murko")
        return doc



import types
from functools import partial
from unittest.mock import MagicMock, call, patch

import bluesky.plan_stubs as bps
import numpy as np
import pytest
from bluesky.run_engine import RunEngine, RunEngineResult
from bluesky.simulators import assert_message_and_return_remaining
from bluesky.utils import FailedStatus, Msg
from dodal.beamlines import i03
from dodal.devices.detector.det_dim_constants import (
    EIGER_TYPE_EIGER2_X_16M,
)
from dodal.devices.fast_grid_scan import ZebraFastGridScan
from dodal.devices.smargon import CombinedMove
from dodal.devices.synchrotron import SynchrotronMode
from dodal.devices.zocalo import ZocaloStartInfo
from numpy import isclose
from ophyd.sim import NullStatus
from ophyd.status import Status
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan import (
    BeamlineSpecificFGSFeatures,
    FlyScanEssentialDevices,
    _fetch_xrc_results_from_zocalo,
    common_flyscan_xray_centre,
    kickoff_and_complete_gridscan,
    run_gridscan,
    wait_for_gridscan_valid,
)
from mx_bluesky.common.experiment_plans.inner_plans.read_hardware import (
    read_hardware_plan,
)
from mx_bluesky.common.external_interaction.callbacks.common.logging_callback import (
    VerbosePlanExecutionLoggingCallback,
)
from mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback import (
    ZocaloCallback,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback import (
    GridscanISPyBCallback,
    ispyb_activation_wrapper,
)
from mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback import (
    GridscanNexusFileCallback,
)
from mx_bluesky.common.external_interaction.ispyb.ispyb_store import (
    IspybIds,
)
from mx_bluesky.common.parameters.constants import DocDescriptorNames
from mx_bluesky.common.parameters.gridscan import SpecifiedThreeDGridScan
from mx_bluesky.common.utils.exceptions import (
    CrystalNotFoundException,
    WarningException,
)
from mx_bluesky.common.xrc_result import XRayCentreEventHandler, XRayCentreResult
from tests.conftest import (
    RunEngineSimulator,
    create_dummy_scan_spec,
)
from tests.unit_tests.hyperion.experiment_plans.conftest import mock_zocalo_trigger

from ....conftest import TestData
from ...conftest import (
    create_gridscan_callbacks,
    modified_store_grid_scan_mock,
    run_generic_ispyb_handler_setup,
)

ReWithSubs = tuple[RunEngine, tuple[GridscanNexusFileCallback, GridscanISPyBCallback]]


class CompleteException(Exception):
    # To avoid having to run through the entire plan during tests
    pass


def mock_plan():
    yield from bps.null()


@patch(
    "mx_bluesky.common.external_interaction.callbacks.xray_centre.ispyb_callback.StoreInIspyb",
    modified_store_grid_scan_mock,
)
class TestFlyscanXrayCentrePlan:
    td: TestData = TestData()

    def test_eiger2_x_16_detector_specified(
        self,
        test_fgs_params: SpecifiedThreeDGridScan,
    ):
        assert (
            test_fgs_params.detector_params.detector_size_constants.det_type_string
            == EIGER_TYPE_EIGER2_X_16M
        )

    def test_when_run_gridscan_called_then_generator_returned(
        self,
    ):
        plan = run_gridscan(MagicMock(), MagicMock(), MagicMock())
        assert isinstance(plan, types.GeneratorType)

    def test_when_run_gridscan_called_ispyb_deposition_made_and_records_errors(
        self,
        RE: RunEngine,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        ispyb_callback = GridscanISPyBCallback(param_type=SpecifiedThreeDGridScan)
        RE.subscribe(ispyb_callback)

        error = None
        with patch.object(fake_fgs_composite.smargon.omega, "set") as mock_set:
            error = AssertionError("Test Exception")
            mock_set.return_value = FailedStatus(error)
            with pytest.raises(FailedStatus) as exc:
                RE(
                    ispyb_activation_wrapper(
                        common_flyscan_xray_centre(
                            fake_fgs_composite, test_fgs_params, beamline_specific
                        ),
                        test_fgs_params,
                    ),
                )

        assert exc.value.args[0] is error
        ispyb_callback.ispyb.end_deposition.assert_called_once_with(  # type: ignore
            IspybIds(data_collection_group_id=0, data_collection_ids=(0, 0)),
            "fail",
            "Test Exception",
        )

    @patch("bluesky.plan_stubs.abs_set", autospec=True)
    def test_results_passed_to_move_motors(
        self,
        bps_abs_set: MagicMock,
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
        RE: RunEngine,
    ):
        from mx_bluesky.common.device_setup_plans.manipulate_sample import move_x_y_z

        motor_position = test_fgs_params.FGS_params.grid_position_to_motor_position(
            np.array([1, 2, 3])
        )
        RE(move_x_y_z(fake_fgs_composite.smargon, *motor_position))
        bps_abs_set.assert_called_with(
            fake_fgs_composite.smargon,
            CombinedMove(x=motor_position[0], y=motor_position[1], z=motor_position[2]),
            group="move_x_y_z",
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
    )
    def test_individual_plans_triggered_once_and_only_once_in_composite_run(
        self,
        zoc_trigger: MagicMock,
        run_gridscan: MagicMock,
        RE_with_subs: ReWithSubs,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        RE, _ = RE_with_subs

        def wrapped_gridscan_and_move():
            yield from common_flyscan_xray_centre(
                fake_fgs_composite,
                test_fgs_params,
                beamline_specific,
            )

        RE(wrapped_gridscan_and_move())
        run_gridscan.assert_called_once()
        beamline_specific.setup_trigger_plan.assert_called_once()  # type: ignore
        beamline_specific.tidy_plan.assert_called_once()  # type: ignore

    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
    )
    def test_waits_for_motion_program(
        self,
        check_topup_and_wait,
        RE: RunEngine,
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
        done_status: Status,
    ):
        fake_fgs_composite.eiger.unstage = MagicMock(return_value=done_status)
        fgs = i03.zebra_fast_grid_scan(connect_immediately=True, mock=True)
        fgs.KICKOFF_TIMEOUT = 0.1
        fgs.complete = MagicMock(return_value=done_status)
        set_mock_value(fgs.motion_program.running, 1)

        def test_plan():
            yield from kickoff_and_complete_gridscan(
                fgs,
                fake_fgs_composite.eiger,
                fake_fgs_composite.synchrotron,
                [
                    test_fgs_params.scan_points_first_grid,
                    test_fgs_params.scan_points_second_grid,
                ],
                test_fgs_params.scan_indices,
            )

        with pytest.raises(FailedStatus):
            RE(test_plan())
        fgs.KICKOFF_TIMEOUT = 1
        set_mock_value(fgs.motion_program.running, 0)
        set_mock_value(fgs.status, 1)
        res = RE(test_plan())

        assert isinstance(res, RunEngineResult)
        assert res.exit_status == "success"

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.sleep",
        autospec=True,
    )
    def test_GIVEN_scan_already_valid_THEN_wait_for_GRIDSCAN_returns_immediately(
        self, patch_sleep: MagicMock, RE: RunEngine
    ):
        test_fgs: ZebraFastGridScan = i03.zebra_fast_grid_scan(
            connect_immediately=True, mock=True
        )

        set_mock_value(test_fgs.position_counter, 0)
        set_mock_value(test_fgs.scan_invalid, False)

        RE(wait_for_gridscan_valid(test_fgs))

        patch_sleep.assert_not_called()

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.sleep",
        autospec=True,
    )
    def test_GIVEN_scan_not_valid_THEN_wait_for_GRIDSCAN_raises_and_sleeps_called(
        self, patch_sleep: MagicMock, RE: RunEngine
    ):
        test_fgs: ZebraFastGridScan = i03.zebra_fast_grid_scan(
            connect_immediately=True, mock=True
        )

        set_mock_value(test_fgs.scan_invalid, True)
        set_mock_value(test_fgs.position_counter, 0)

        with pytest.raises(WarningException):
            RE(wait_for_gridscan_valid(test_fgs))

        patch_sleep.assert_called()

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.abs_set",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.mv",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.wait_for_gridscan_valid",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.external_interaction.nexus.write_nexus.NexusWriter",
        autospec=True,
        spec_set=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_when_grid_scan_ran_then_eiger_disarmed_before_zocalo_end(
        self,
        mock_check_topup,
        nexuswriter,
        wait_for_valid,
        mock_mv,
        mock_complete,
        mock_kickoff,
        mock_abs_set,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        RE_with_subs: ReWithSubs,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        test_fgs_params.x_steps = 9
        test_fgs_params.y_steps = 10
        test_fgs_params.z_steps = 12
        RE, (nexus_cb, ispyb_cb) = RE_with_subs
        # Put both mocks in a parent to easily capture order
        mock_parent = MagicMock()
        fake_fgs_composite.eiger.disarm_detector = mock_parent.disarm
        assert isinstance(ispyb_cb.emit_cb, ZocaloCallback)
        ispyb_cb.emit_cb.zocalo_interactor.run_end = mock_parent.run_end

        fake_fgs_composite.eiger.filewriters_finished = NullStatus()  # type: ignore
        fake_fgs_composite.eiger.odin.check_and_wait_for_odin_state = MagicMock(
            return_value=True
        )
        fake_fgs_composite.eiger.odin.file_writer.num_captured.sim_put(1200)  # type: ignore
        fake_fgs_composite.eiger.stage = MagicMock(
            return_value=Status(None, None, 0, True, True)
        )

        with patch(
            "mx_bluesky.common.external_interaction.callbacks.xray_centre.nexus_callback.NexusWriter.create_nexus_file",
            autospec=True,
        ):
            [RE.subscribe(cb) for cb in (nexus_cb, ispyb_cb)]
            RE(
                ispyb_activation_wrapper(
                    common_flyscan_xray_centre(
                        fake_fgs_composite, test_fgs_params, beamline_specific
                    ),
                    test_fgs_params,
                )
            )

        mock_parent.assert_has_calls([call.disarm(), call.run_end(0), call.run_end(0)])

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.wait",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_fgs_arms_eiger_without_grid_detect(
        self,
        mock_topup,
        mock_kickoff,
        mock_complete,
        mock_wait,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        RE: RunEngine,
        done_status: Status,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        fake_fgs_composite.eiger.unstage = MagicMock(return_value=done_status)
        RE(run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific))
        fake_fgs_composite.eiger.stage.assert_called_once()  # type: ignore
        fake_fgs_composite.eiger.unstage.assert_called_once()

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.wait",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_when_grid_scan_fails_with_exception_then_detector_disarmed_and_correct_exception_returned(
        self,
        mock_topup,
        mock_complete,
        mock_wait,
        mock_kickoff,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        RE: RunEngine,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        beamline_specific.read_pre_flyscan_plan = partial(
            read_hardware_plan,
            [],
            DocDescriptorNames.HARDWARE_READ_DURING,
        )

        mock_complete.side_effect = CompleteException()

        fake_fgs_composite.eiger.stage = MagicMock(
            return_value=Status(None, None, 0, True, True)
        )

        fake_fgs_composite.eiger.filewriters_finished = NullStatus()

        fake_fgs_composite.eiger.odin.check_and_wait_for_odin_state = MagicMock()

        fake_fgs_composite.eiger.disarm_detector = MagicMock()
        fake_fgs_composite.eiger.disable_roi_mode = MagicMock()

        with pytest.raises(CompleteException):
            RE(run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific))

        fake_fgs_composite.eiger.disable_roi_mode.assert_called()
        fake_fgs_composite.eiger.disarm_detector.assert_called()

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.kickoff",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.bps.complete",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.external_interaction.callbacks.common.zocalo_callback.ZocaloTrigger",
        autospec=True,
    )
    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        autospec=True,
    )
    def test_kickoff_and_complete_gridscan_triggers_zocalo(
        self,
        mock_topup,
        mock_zocalo_trigger_class: MagicMock,
        mock_complete: MagicMock,
        mock_kickoff: MagicMock,
        RE: RunEngine,
        fake_fgs_composite: FlyScanEssentialDevices,
        dummy_rotation_data_collection_group_info,
        zebra_fast_grid_scan: ZebraFastGridScan,
    ):
        id_1, id_2 = 100, 200

        _, ispyb_cb = create_gridscan_callbacks()
        ispyb_cb.active = True
        ispyb_cb.ispyb = MagicMock()
        ispyb_cb.params = MagicMock()
        ispyb_cb.ispyb_ids.data_collection_ids = (id_1, id_2)
        ispyb_cb.data_collection_group_info = dummy_rotation_data_collection_group_info
        assert isinstance(ispyb_cb.emit_cb, ZocaloCallback)

        mock_zocalo_trigger = ispyb_cb.emit_cb.zocalo_interactor

        fake_fgs_composite.eiger.unstage = MagicMock()
        fake_fgs_composite.eiger.odin.file_writer.id.sim_put("test/filename")  # type: ignore

        x_steps, y_steps, z_steps = 10, 20, 30

        RE.subscribe(ispyb_cb)

        RE(
            kickoff_and_complete_gridscan(
                zebra_fast_grid_scan,
                fake_fgs_composite.eiger,
                fake_fgs_composite.synchrotron,
                scan_points=create_dummy_scan_spec(x_steps, y_steps, z_steps),
                scan_start_indices=[0, x_steps * y_steps],
            )
        )

        expected_start_infos = [
            ZocaloStartInfo(id_1, "test/filename", 0, x_steps * y_steps, 0),
            ZocaloStartInfo(
                id_2, "test/filename", x_steps * y_steps, x_steps * z_steps, 1
            ),
        ]

        expected_start_calls = [
            call(expected_start_infos[0]),
            call(expected_start_infos[1]),
        ]

        assert mock_zocalo_trigger.run_start.call_count == 2  # type: ignore
        assert mock_zocalo_trigger.run_start.mock_calls == expected_start_calls  # type: ignore

        assert mock_zocalo_trigger.run_end.call_count == 2  # type: ignore
        assert mock_zocalo_trigger.run_end.mock_calls == [call(id_1), call(id_2)]  # type: ignore

    @patch(
        "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
        new=MagicMock(side_effect=lambda *_, **__: iter([Msg("check_topup")])),
    )
    def test_read_hardware_during_collection_occurs_after_eiger_arm(
        self,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        sim_run_engine: RunEngineSimulator,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        beamline_specific.read_during_collection_plan = partial(
            read_hardware_plan,
            [fake_fgs_composite.eiger.bit_depth],  # type: ignore # see https://github.com/bluesky/bluesky/issues/1809
            DocDescriptorNames.HARDWARE_READ_DURING,
        )
        sim_run_engine.add_handler(
            "read",
            lambda msg: {"values": {"value": SynchrotronMode.USER}},
            "synchrotron-synchrotron_mode",
        )
        msgs = sim_run_engine.simulate_plan(
            run_gridscan(fake_fgs_composite, test_fgs_params, beamline_specific)
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "stage" and msg.obj.name == "eiger"
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "kickoff"
            and msg.obj == beamline_specific.fgs_motors,
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "create"
        )
        msgs = assert_message_and_return_remaining(
            msgs,
            lambda msg: msg.command == "read" and msg.obj.name == "eiger_bit_depth",
        )
        msgs = assert_message_and_return_remaining(
            msgs, lambda msg: msg.command == "save"
        )

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        autospec=True,
    )
    def test_when_gridscan_succeeds_and_results_fetched_ispyb_comment_appended_to(
        self,
        run_gridscan: MagicMock,
        RE_with_subs: ReWithSubs,
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        RE, (nexus_cb, ispyb_cb) = RE_with_subs

        def _wrapped_gridscan_and_move():
            run_generic_ispyb_handler_setup(ispyb_cb, test_fgs_params)
            yield from common_flyscan_xray_centre(
                fake_fgs_composite,
                test_fgs_params,
                beamline_specific,
            )

        RE.subscribe(VerbosePlanExecutionLoggingCallback())
        beamline_specific.get_xrc_results_from_zocalo = True
        RE(ispyb_activation_wrapper(_wrapped_gridscan_and_move(), test_fgs_params))
        app_to_comment: MagicMock = ispyb_cb.ispyb.append_to_comment  # type:ignore
        app_to_comment.assert_called()
        append_aperture_call = app_to_comment.call_args_list[0].args[1]
        assert "Aperture:" in append_aperture_call

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        autospec=True,
    )
    async def test_results_adjusted_and_event_raised(
        self,
        run_gridscan: MagicMock,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        beamline_specific: BeamlineSpecificFGSFeatures,
        RE_with_subs: ReWithSubs,
    ):
        RE, _ = RE_with_subs
        beamline_specific.get_xrc_results_from_zocalo = True
        x_ray_centre_event_handler = XRayCentreEventHandler()
        RE.subscribe(x_ray_centre_event_handler)
        mock_zocalo_trigger(fake_fgs_composite.zocalo, TestData.test_result_large)

        def plan():
            yield from _fetch_xrc_results_from_zocalo(
                fake_fgs_composite.zocalo, test_fgs_params
            )

        RE(plan())

        actual = x_ray_centre_event_handler.xray_centre_results
        expected = XRayCentreResult(
            centre_of_mass_mm=np.array([0.05, 0.15, 0.25]),
            bounding_box_mm=(
                np.array([0.15, 0.15, 0.15]),
                np.array([0.75, 0.75, 0.65]),
            ),
            max_count=105062,
            total_count=2387574,
            sample_id=12345,
        )
        assert actual and len(actual) == 1
        assert all(isclose(actual[0].centre_of_mass_mm, expected.centre_of_mass_mm))
        assert all(isclose(actual[0].bounding_box_mm[0], expected.bounding_box_mm[0]))
        assert all(isclose(actual[0].bounding_box_mm[1], expected.bounding_box_mm[1]))

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.kickoff_and_complete_gridscan",
        MagicMock(),
    )
    def test_run_gridscan_and_fetch_results_discards_results_below_threshold(
        self,
        fake_fgs_composite: FlyScanEssentialDevices,
        test_fgs_params: SpecifiedThreeDGridScan,
        beamline_specific: BeamlineSpecificFGSFeatures,
        RE: RunEngine,
    ):
        beamline_specific.get_xrc_results_from_zocalo = True
        callback = XRayCentreEventHandler()
        RE.subscribe(callback)

        mock_zocalo_trigger(
            fake_fgs_composite.zocalo,
            TestData.test_result_medium
            + TestData.test_result_below_threshold
            + TestData.test_result_small,
        )
        RE(_fetch_xrc_results_from_zocalo(fake_fgs_composite.zocalo, test_fgs_params))

        assert callback.xray_centre_results and len(callback.xray_centre_results) == 2
        assert [r.max_count for r in callback.xray_centre_results] == [50000, 1000]

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        autospec=True,
    )
    def test_when_gridscan_finds_no_xtal_exception_is_raised(
        self,
        run_gridscan: MagicMock,
        RE_with_subs: ReWithSubs,
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
    ):
        RE, (nexus_cb, ispyb_cb) = RE_with_subs
        beamline_specific.get_xrc_results_from_zocalo = True

        def wrapped_gridscan_and_move():
            run_generic_ispyb_handler_setup(ispyb_cb, test_fgs_params)
            yield from common_flyscan_xray_centre(
                fake_fgs_composite,
                test_fgs_params,
                beamline_specific,
            )

        mock_zocalo_trigger(fake_fgs_composite.zocalo, [])
        with pytest.raises(CrystalNotFoundException):
            RE(ispyb_activation_wrapper(wrapped_gridscan_and_move(), test_fgs_params))

    @patch(
        "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
        MagicMock(),
    )
    def test_dummy_result_returned_when_gridscan_finds_no_xtal_and_commissioning_mode_enabled(
        self,
        RE: RunEngine,
        test_fgs_params: SpecifiedThreeDGridScan,
        fake_fgs_composite: FlyScanEssentialDevices,
        beamline_specific: BeamlineSpecificFGSFeatures,
        baton_in_commissioning_mode,
    ):
        xrc_event_handler = XRayCentreEventHandler()
        RE.subscribe(xrc_event_handler)
        beamline_specific.get_xrc_results_from_zocalo = True

        mock_zocalo_trigger(fake_fgs_composite.zocalo, [])
        RE(
            common_flyscan_xray_centre(
                fake_fgs_composite,
                test_fgs_params,
                beamline_specific,
            )
        )

        results = xrc_event_handler.xray_centre_results or []
        assert len(results) == 1
        result = results[0]
        assert result.sample_id == test_fgs_params.sample_id
        assert result.max_count == 10000
        assert result.total_count == 100000
        assert all(np.isclose(result.bounding_box_mm[0], [1.95, 0.95, 0.45]))
        assert all(np.isclose(result.bounding_box_mm[1], [2.05, 1.05, 0.55]))
        assert all(np.isclose(result.centre_of_mass_mm, [2.0, 1.0, 0.5]))



import gc
import os
from dataclasses import fields
from unittest.mock import patch
from weakref import WeakValueDictionary

import pytest
from blueapi.core import BlueskyContext
from bluesky import RunEngine
from ophyd_async.core import Device
from ophyd_async.plan_stubs import ensure_connected

from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.hyperion.baton_handler import _initialise_udc
from mx_bluesky.hyperion.experiment_plans.load_centre_collect_full_plan import (
    LoadCentreCollectComposite,
)
from mx_bluesky.hyperion.utils.context import setup_context

weak_ids_to_devices = WeakValueDictionary()


MAX_DEVICE_COUNT = 4


@pytest.fixture
def patch_setup_devices(request):
    dev_mode = request.param
    from mx_bluesky.hyperion.utils.context import setup_devices

    def patched_setup_devices(context: BlueskyContext, _: bool):
        setup_devices(context, dev_mode)

    with patch(
        "mx_bluesky.hyperion.baton_handler.setup_devices",
        side_effect=patched_setup_devices,
    ) as patched_func:
        yield patched_func


@pytest.fixture
def patch_ensure_connected():
    unpatched = ensure_connected

    def patched_func(*devices: Device, mock=False, timeout=1, force_reconnect=False):
        timeout = 1
        yield from unpatched(
            *devices, mock=mock, timeout=timeout, force_reconnect=force_reconnect
        )

    with patch(
        "blueapi.utils.connect_devices.ensure_connected", side_effect=patched_func
    ) as p:
        yield p


@pytest.mark.parametrize(
    "i, patch_setup_devices",
    [[i, True] for i in range(1, 101)],
    indirect=["patch_setup_devices"],
)
@pytest.mark.system_test
@patch.dict(os.environ, {"BEAMLINE": "i03"})
def test_udc_reloads_all_devices_soak_test_dev_mode(
    RE: RunEngine, i: int, patch_setup_devices
):
    reinitialise_beamline(True, i)


@pytest.mark.parametrize(
    "i, patch_setup_devices",
    [[i, False] for i in range(1, 101)],
    indirect=["patch_setup_devices"],
)
@patch.dict(os.environ, {"BEAMLINE": "i03"})
@patch("ophyd_async.plan_stubs._ensure_connected.DEFAULT_TIMEOUT", 1)
@pytest.mark.timeout(10)
def test_udc_reloads_all_devices_soak_test_real(
    RE: RunEngine, i: int, patch_setup_devices, patch_ensure_connected
):
    """
    Deliberately not part of main system tests because this is SLOW and requires
    a beamline to connect to.
    """
    reinitialise_beamline(False, i)


def reinitialise_beamline(dev_mode: bool, i: int):
    """Reinitialise the beamline, which should cause all beamline devices
    to be reinstantiated, and then check for memory leaks (potentially caused
    if something holds a reference to the device).
    Since GC is in general implementation dependent and non-deterministic,
    it is not guaranteed that devices will be collected after only one iteration.
    Therefore this test is run a number of times in a loop and we check that the
    outstanding instances do not exceed some threshold value."""

    context = setup_context(dev_mode)
    devices_before_reset: LoadCentreCollectComposite = device_composite_from_context(
        context, LoadCentreCollectComposite
    )
    for f in fields(devices_before_reset):
        device = getattr(devices_before_reset, f.name)
        weak_ids_to_devices[id(device)] = device
    _initialise_udc(context, dev_mode)
    devices_after_reset: LoadCentreCollectComposite = device_composite_from_context(
        context, LoadCentreCollectComposite
    )
    for f in fields(devices_after_reset):
        device = getattr(devices_after_reset, f.name)
        weak_ids_to_devices[id(device)] = device
    for f in fields(devices_after_reset):
        device_after_reset = getattr(devices_after_reset, f.name)
        device_before_reset = getattr(devices_before_reset, f.name)
        assert device_before_reset is not device_after_reset, (
            f"{id(device_before_reset)} == {id(device_after_reset)}"
        )
    check_instances_are_garbage_collected(i)
    context.run_engine.loop.call_soon_threadsafe(context.run_engine.loop.stop)


def check_instances_are_garbage_collected(i: int):
    gc.collect()
    device_counts: dict[str, int] = {}
    for ref in weak_ids_to_devices.valuerefs():
        device = ref()
        if device is not None:
            device_counts[device.name] = device_counts.get(device.name, 0) + 1

    devices_by_count = sorted([(count, name) for name, count in device_counts.items()])
    print(
        f"Dictionary size is {len(weak_ids_to_devices)}, total live references is "
        f"{sum(device_and_count[0] for device_and_count in devices_by_count)}"
    )
    print(
        f"Max count device is {devices_by_count[-1]}, min count device is {devices_by_count[0]}"
    )
    for name, count in device_counts.items():
        max_count = min(MAX_DEVICE_COUNT, i * 2)
        assert count <= max_count, (
            f"Device count {name} exceeded max expected references {count}"
        )



from unittest.mock import patch

import pytest

from mx_bluesky.common.parameters.constants import _get_oav_config_json_path


@pytest.mark.parametrize(
    "beamline, test_mode, expected_path",
    [
        ("i03", True, "tests/test_data/test_OAVCentring.json"),
        (
            "i03",
            False,
            "/dls_sw/i03/software/daq_configuration/json/OAVCentring_hyperion.json",
        ),
        ("i04", False, "/dls_sw/i04/software/daq_configuration/json/OAVCentring.json"),
    ],
)
def test_get_oav_config_json_path(beamline: str, test_mode: bool, expected_path: str):
    with (
        patch("mx_bluesky.common.parameters.constants.BEAMLINE", new=beamline),
        patch("mx_bluesky.common.parameters.constants.TEST_MODE", new=test_mode),
    ):
        assert _get_oav_config_json_path() == expected_path



from mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan import (
    i04_grid_detect_then_xray_centre,
)
from mx_bluesky.beamlines.i04.thawing_plan import (
    thaw,
    thaw_and_murko_centre,
    thaw_and_stream_to_redis,
)

__all__ = [
    "thaw",
    "thaw_and_stream_to_redis",
    "i04_grid_detect_then_xray_centre",
    "thaw_and_murko_centre",
]



import logging
from typing import cast

from dodal.log import LOGGER
from dodal.utils import get_beamline_name

from mx_bluesky.common.external_interaction.alerting import Metadata
from mx_bluesky.common.external_interaction.alerting._service import (
    ExtraMetadata,
    graylog_url,
    ispyb_url,
)
from mx_bluesky.common.external_interaction.callbacks.common.ispyb_mapping import (
    get_proposal_and_session_from_visit_string,
)


class LoggingAlertService:
    """
    Implement an alert service that raises alerts by generating a specially formatted
    log message, that may be intercepted by a logging service such as graylog and
    used to dispatch the alert.
    """

    def __init__(self, graylog_stream_id: str, level=logging.WARNING):
        """
        Create a new instance of the service
        Args:
            level: The python logging level at which to generate the message
        """
        super().__init__()
        self._level = level
        self._graylog_stream_id = graylog_stream_id

    def _append_extra_metadata(self, metadata: dict[Metadata, str]) -> dict[str, str]:
        with_extras = cast(dict, metadata.copy())
        with_extras[ExtraMetadata.GRAYLOG_URL] = graylog_url(self._graylog_stream_id)
        with_extras[ExtraMetadata.BEAMLINE] = get_beamline_name("")
        if sample_id := metadata.get(Metadata.SAMPLE_ID, None):
            with_extras[ExtraMetadata.ISPYB_URL] = ispyb_url(sample_id)
        if visit := metadata.get(Metadata.VISIT, None):
            proposal, _ = get_proposal_and_session_from_visit_string(visit)
            with_extras[ExtraMetadata.PROPOSAL] = proposal
        return with_extras

    def raise_alert(self, summary: str, content: str, metadata: dict[Metadata, str]):
        message = f"***ALERT*** summary={summary} content={content}"
        with_extras = self._append_extra_metadata(metadata)
        LOGGER.log(
            self._level,
            message,
            extra={
                ExtraMetadata.ALERT_SUMMARY: summary,
                ExtraMetadata.ALERT_CONTENT: content,
            }
            | with_extras,
        )



# from collections.abc import Sequence
from typing import Literal

import bluesky.plan_stubs as bps
import bluesky.preprocessors as bpp
from bluesky.utils import MsgGenerator
from dodal.beamlines import i24
from dodal.common import inject
from dodal.devices.hutch_shutter import HutchShutter
from dodal.devices.i24.aperture import Aperture
from dodal.devices.i24.beam_center import DetectorBeamCenter
from dodal.devices.i24.beamstop import Beamstop
from dodal.devices.i24.dcm import DCM
from dodal.devices.i24.dual_backlight import BacklightPositions, DualBacklight
from dodal.devices.i24.focus_mirrors import FocusMirrorsMode
from dodal.devices.i24.pmac import PMAC
from dodal.devices.motors import YZStage
from dodal.devices.oav.oav_detector import OAVBeamCentreFile
from dodal.devices.zebra.zebra import Zebra

from mx_bluesky.beamlines.i24.serial.dcid import DCID
from mx_bluesky.beamlines.i24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Collect_py3v1 import (
    run_plan_in_wrapper,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_Chip_Manager_py3v1 import (
    upload_chip_map_to_geobrick,
)
from mx_bluesky.beamlines.i24.serial.fixed_target.i24ssx_moveonclick import (
    _move_on_mouse_click_plan,
)
from mx_bluesky.beamlines.i24.serial.log import (
    SSX_LOGGER,
    _read_visit_directory_from_file,
)
from mx_bluesky.beamlines.i24.serial.parameters import (
    FixedTargetParameters,
    get_chip_format,
)
from mx_bluesky.beamlines.i24.serial.parameters.utils import EmptyMapError
from mx_bluesky.beamlines.i24.serial.setup_beamline import pv
from mx_bluesky.beamlines.i24.serial.setup_beamline.ca import caput
from mx_bluesky.beamlines.i24.serial.setup_beamline.pv_abstract import Eiger
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector import (
    _move_detector_stage,
    get_detector_type,
)


@bpp.run_decorator()
def gui_move_backlight(
    position: str, backlight: DualBacklight = inject("backlight")
) -> MsgGenerator:
    bl_pos = BacklightPositions(position)
    yield from bps.abs_set(backlight, bl_pos, wait=True)
    SSX_LOGGER.debug(f"Backlight moved to {bl_pos.value}")


@bpp.run_decorator()
def gui_set_zoom_level(
    position: str, oav: OAVBeamCentreFile = inject("oav")
) -> MsgGenerator:
    yield from bps.abs_set(oav.zoom_controller, position, wait=True)
    SSX_LOGGER.debug(f"Setting zoom level to {position}")


@bpp.run_decorator()
def gui_stage_move_on_click(
    position_px: tuple[int, int],
    oav: OAVBeamCentreFile = inject("oav"),
    pmac: PMAC = inject("pmac"),
) -> MsgGenerator:
    yield from _move_on_mouse_click_plan(oav, pmac, position_px)


@bpp.run_decorator()
def gui_gonio_move_on_click(position_px: tuple[int, int]) -> MsgGenerator:
    oav = i24.oav()
    gonio = i24.vgonio()

    x_pixels_per_micron = yield from bps.rd(oav.microns_per_pixel_x)
    y_pixels_per_micron = yield from bps.rd(oav.microns_per_pixel_y)

    x_um = position_px[0] * x_pixels_per_micron
    y_um = position_px[1] * y_pixels_per_micron

    # gonio is in mm?
    yield from bps.mv(gonio.x, x_um / 1000, gonio.yh, y_um / 1000)  # type: ignore


# See https://github.com/DiamondLightSource/mx-bluesky/issues/853
@bpp.run_decorator()
def gui_sleep(sec: int) -> MsgGenerator:
    for _ in range(sec):
        yield from bps.sleep(1)


@bpp.run_decorator()
def gui_move_detector(
    det: Literal["eiger"],
    detector_stage: YZStage = inject("detector_motion"),
) -> MsgGenerator:
    det_y_target = Eiger.det_y_target
    yield from _move_detector_stage(detector_stage, det_y_target)
    # Make the output readable
    SSX_LOGGER.debug(f"Detector move done, resetting general PV to {det}")
    caput(pv.me14e_gp101, det)


@bpp.run_decorator()
def gui_run_chip_collection(
    sub_dir: str,
    chip_name: str,
    exp_time: float,
    det_dist: float,
    transmission: float,
    n_shots: int,
    chip_type: str,
    map_type: str,
    chip_format: list[int | float],  # for Lite Oxford it's the chipmap
    checker_pattern: bool,
    pump_probe: str,
    laser_dwell: float,
    laser_delay: float,
    pre_pump: float,
    pmac: PMAC = inject("pmac"),
    zebra: Zebra = inject("zebra"),
    aperture: Aperture = inject("aperture"),
    backlight: DualBacklight = inject("backlight"),
    beamstop: Beamstop = inject("beamstop"),
    detector_stage: YZStage = inject("detector_motion"),
    shutter: HutchShutter = inject("shutter"),
    dcm: DCM = inject("dcm"),
    mirrors: FocusMirrorsMode = inject("focus_mirrors"),
    beam_center_eiger: DetectorBeamCenter = inject("eiger_bc"),
) -> MsgGenerator:
    """Set the parameter model for the data collection.

    Args:
        sub_dir (str): subdirectory of the visit to write data in.
        chip_name (str): a name identifying the current chip collection, will be used
            as filename.
        exp_time (float): exposure time of each window shot, in s.
        det_dist (float): sample-detector distance, in mm.
        transmission (float): requested beam intensity transmission, expressed as
            a fraction, e.g. 0.3.
        n_shots (int): number of times each window should be collected.
        chip_type (str): type of chip in use.
        map_type (str): if an Oxford chip is used, define whether it's a full chip
            collection or lite mapping is in use. For all other chip, this will be None.
        chip_format (list[int|float]): for a custom chip, a list of the number of x,y
            steps and the x,y step size. For an Oxford chip, the list should be empty
            if collecting a full chip and a list of the block numbers to scan for a
            lite collection.
        checker_pattern (bool): whether checker_pattern is turned on, ie. only every
            other window in a block gets collected
        pump_probe (str): pump probe setting.
        laser_dwell (float): laser exposure time for pump probe collections, in s.
        laser_delay (float): delay between laser exposure and collection, in s.
        pre_pump (float): pre-pump exposure time for a pump probe short2 collection,
            ie a pump-in-probe where the collection starts during the pump.
    """
    # NOTE still a work in progress, adding to it as the ui grows
    # See progression of https://github.com/DiamondLightSource/mx-daq-ui/issues/3
    det_type = yield from get_detector_type(detector_stage)
    _format = chip_format if ChipType[chip_type] is ChipType.Custom else None
    chip_params = get_chip_format(ChipType[chip_type], _format)
    if ChipType[chip_type] in [ChipType.Oxford, ChipType.OxfordInner]:
        mapping = MappingType.Lite if map_type == "Lite" else MappingType.NoMap
        if mapping is MappingType.Lite and len(chip_format) == 0:
            # this logic should go in the gui with error message.
            raise EmptyMapError("No blocks chosen")
        chip_map = chip_format
    else:
        mapping = MappingType.NoMap
        chip_map = []

    params = {
        "visit": _read_visit_directory_from_file().as_posix(),  # noqa
        "directory": sub_dir,
        "filename": chip_name,
        "exposure_time_s": exp_time,
        "detector_distance_mm": det_dist,
        "detector_name": str(det_type),
        "num_exposures": n_shots,
        "transmission": transmission,
        "chip": chip_params,
        "map_type": mapping,
        "chip_map": chip_map,
        "pump_repeat": PumpProbeSetting[pump_probe],  # pump_repeat,
        "laser_dwell_s": laser_dwell,
        "laser_delay_s": laser_delay,
        "checker_pattern": checker_pattern,
        "pre_pump_exposure_s": pre_pump,
    }

    parameters = FixedTargetParameters(**params)

    # Create collection directory
    parameters.collection_directory.mkdir(parents=True, exist_ok=True)

    if parameters.chip_map:
        yield from upload_chip_map_to_geobrick(pmac, parameters.chip_map)

    beam_center_device = beam_center_eiger
    SSX_LOGGER.info("Beam center device ready")

    # DCID instance - do not create yet
    dcid = DCID(emit_errors=False, expt_params=parameters)  # noqa
    SSX_LOGGER.info("DCID created")

    yield from run_plan_in_wrapper(
        zebra,
        pmac,
        aperture,
        backlight,
        beamstop,
        detector_stage,
        shutter,
        dcm,
        mirrors,
        beam_center_device,
        parameters,
        dcid,
    )



from blueapi.core import BlueskyContext
from dodal.common.beamlines.beamline_utils import clear_devices
from dodal.utils import collect_factories, get_beamline_based_on_environment_variable

import mx_bluesky.hyperion.experiment_plans as hyperion_plans
from mx_bluesky.common.utils.log import LOGGER


def setup_context(dev_mode: bool = False) -> BlueskyContext:
    context = BlueskyContext()
    context.with_plan_module(hyperion_plans)

    setup_devices(context, dev_mode)

    LOGGER.info(f"Plans found in context: {context.plan_functions.keys()}")

    return context


def clear_all_device_caches(context: BlueskyContext):
    context.unregister_all_devices()
    clear_devices()

    for f in collect_factories(get_beamline_based_on_environment_variable()).values():
        if hasattr(f, "cache_clear"):
            f.cache_clear()  # type: ignore


def setup_devices(context: BlueskyContext, dev_mode: bool):
    context.with_dodal_module(
        get_beamline_based_on_environment_variable(),
        mock=dev_mode,
    )



import os
from datetime import datetime
from logging import INFO, WARNING
from unittest.mock import MagicMock, patch

import pytest

from mx_bluesky.common.external_interaction.alerting import (
    Metadata,
    get_alerting_service,
    set_alerting_service,
)
from mx_bluesky.common.external_interaction.alerting._service import ExtraMetadata
from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.hyperion.parameters.constants import CONST

EXPECTED_GRAYLOG_URL = (
    "https://graylog.diamond.ac.uk/search?streams=66264f5519ccca6d1c9e4e03&"
    "rangetype=absolute&"
    "from=2025-08-25T15%3A27%3A24%2B00%3A00&"
    "to=2025-08-25T15%3A32%3A25%2B00%3A00"
)


@pytest.fixture(autouse=True)
def fixup_time():
    with patch(
        "mx_bluesky.common.external_interaction.alerting._service.datetime",
        MagicMock(
            **{"now.return_value": datetime.fromisoformat("2025-08-25T15:32:24Z")}  # type: ignore
        ),
    ) as patched_now:
        yield patched_now


@pytest.fixture(autouse=True)
def fixup_beamline():
    with patch.dict(os.environ, {"BEAMLINE": "i03"}):
        yield


@pytest.mark.parametrize("level", [WARNING, INFO])
@patch("mx_bluesky.common.external_interaction.alerting.log_based_service.LOGGER")
def test_logging_alerting_service_raises_a_log_message(mock_logger: MagicMock, level):
    set_alerting_service(LoggingAlertService(CONST.GRAYLOG_STREAM_ID, level))
    get_alerting_service().raise_alert("Test summary", "Test message", {})

    mock_logger.log.assert_called_once_with(
        level,
        "***ALERT*** summary=Test summary content=Test message",
        extra={
            ExtraMetadata.ALERT_SUMMARY: "Test summary",
            ExtraMetadata.ALERT_CONTENT: "Test message",
            ExtraMetadata.BEAMLINE: "i03",
            ExtraMetadata.GRAYLOG_URL: EXPECTED_GRAYLOG_URL,
        },
    )


@patch("mx_bluesky.common.external_interaction.alerting.log_based_service.LOGGER")
def test_logging_alerting_service_raises_a_log_message_with_additional_metadata_when_sample_id_present(
    mock_logger: MagicMock,
):
    set_alerting_service(LoggingAlertService(CONST.GRAYLOG_STREAM_ID, WARNING))
    get_alerting_service().raise_alert(
        "Test summary",
        "Test message",
        {Metadata.SAMPLE_ID: "123456", Metadata.VISIT: "cm14451-2"},
    )

    mock_logger.log.assert_called_once_with(
        WARNING,
        "***ALERT*** summary=Test summary content=Test message",
        extra={
            ExtraMetadata.ALERT_SUMMARY: "Test summary",
            ExtraMetadata.ALERT_CONTENT: "Test message",
            ExtraMetadata.BEAMLINE: "i03",
            ExtraMetadata.GRAYLOG_URL: EXPECTED_GRAYLOG_URL,
            ExtraMetadata.ISPYB_URL: "https://ispyb.diamond.ac.uk/samples/sid/123456",
            Metadata.SAMPLE_ID: "123456",
            ExtraMetadata.PROPOSAL: "cm14451",
            Metadata.VISIT: "cm14451-2",
        },
    )



from unittest.mock import patch

import pytest
from bluesky.run_engine import RunEngine
from dodal.devices.motors import YZStage
from ophyd_async.testing import set_mock_value

from mx_bluesky.beamlines.i24.serial.parameters.constants import SSXType
from mx_bluesky.beamlines.i24.serial.setup_beamline import Eiger
from mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector import (
    EXPT_TYPE_DETECTOR_PVS,
    DetRequest,
    _get_requested_detector,
    get_detector_type,
    setup_detector_stage,
)


def test_get_detector_type(RE, detector_stage: YZStage):
    set_mock_value(detector_stage.y.user_readback, -59)
    det_type = RE(get_detector_type(detector_stage)).plan_result
    assert det_type.name == "eiger"


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector.caget")
def test_get_requested_detector(fake_caget):
    fake_caget.return_value = "0"
    assert _get_requested_detector("some_pv") == Eiger.name


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector.caget")
def test_get_requested_detector_raises_error_for_invalid_value(fake_caget):
    fake_caget.return_value = "something"
    with pytest.raises(ValueError):
        _get_requested_detector("some_pv")


@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector.caget")
@patch("mx_bluesky.beamlines.i24.serial.setup_beamline.setup_detector.caput")
@pytest.mark.parametrize(
    "requested_detector_value, serial_type, detector_target",
    [
        (DetRequest.eiger.value, SSXType.FIXED, Eiger.det_y_target),
    ],
)
async def test_setup_detector_stage(
    fake_caput,
    fake_caget,
    requested_detector_value,
    serial_type,
    detector_target,
    detector_stage: YZStage,
    RE: RunEngine,
):
    fake_caget.return_value = requested_detector_value
    RE(setup_detector_stage(serial_type, detector_stage))
    fake_caput.assert_called_once_with(
        EXPT_TYPE_DETECTOR_PVS[serial_type],
        _get_requested_detector(requested_detector_value),
    )
    assert await detector_stage.y.user_setpoint.get_value() == detector_target



from functools import partial
from unittest.mock import MagicMock, patch

import pytest
from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp
from bluesky.run_engine import RunEngine
from dodal.devices.backlight import Backlight
from dodal.devices.robot import BartRobot
from ophyd_async.testing import set_mock_value

from mx_bluesky.common.external_interaction.alerting import Metadata
from mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy import (
    robot_load_and_snapshots,
)
from mx_bluesky.hyperion.external_interaction.callbacks.alert_on_container_change import (
    AlertOnContainerChange,
)
from mx_bluesky.hyperion.parameters.constants import CONST

TEST_SAMPLE_ID = 10
TEST_VISIT = "cm1234-67"


def wrap_plan(container: int, plan):
    yield from bpp.run_wrapper(
        plan(),
        md={
            "metadata": {
                "container": container,
                "sample_id": TEST_SAMPLE_ID,
                "visit": TEST_VISIT,
            },
            "activate_callbacks": ["AlertOnContainerChange"],
        },
    )


def dummy_plan_with_container(robot: BartRobot, container: int):
    def my_plan():
        yield from bps.create(name=CONST.DESCRIPTORS.ROBOT_PRE_LOAD)
        yield from bps.read(robot)
        yield from bps.save()

    yield from wrap_plan(container, my_plan)


@patch.dict("os.environ", {"BEAMLINE": "i03"})
def test_when_data_collected_on_the_same_container_then_does_not_alert_multiple_times(
    RE: RunEngine, mock_alert_service: MagicMock, robot: BartRobot
):
    RE.subscribe(AlertOnContainerChange())

    set_mock_value(robot.current_puck, 5)

    RE(dummy_plan_with_container(robot, 5))
    RE(dummy_plan_with_container(robot, 5))
    RE(dummy_plan_with_container(robot, 5))

    mock_alert_service.raise_alert.assert_not_called()


@patch.dict("os.environ", {"BEAMLINE": "i03"})
def test_when_data_collected_on_new_container_then_alerts(
    RE: RunEngine,
    mock_alert_service: MagicMock,
    robot: BartRobot,
):
    RE.subscribe(AlertOnContainerChange())
    set_mock_value(robot.current_puck, 5)

    RE(dummy_plan_with_container(robot, 10))

    mock_alert_service.raise_alert.assert_called_once_with(
        "UDC moved on to puck 10 on i03",
        "Hyperion finished container 5 and moved on to 10",
        {
            Metadata.SAMPLE_ID: "10",
            Metadata.VISIT: "cm1234-67",
            Metadata.CONTAINER: "10",
        },
    )


@patch.dict("os.environ", {"BEAMLINE": "i03"})
@patch(
    "mx_bluesky.hyperion.experiment_plans.robot_load_and_change_energy.do_robot_load"
)
def test_robot_load_and_snapshots_triggers_alert(
    patched_robot_load: MagicMock,
    RE: RunEngine,
    mock_alert_service: MagicMock,
    robot: BartRobot,
    backlight: Backlight,
):
    RE.subscribe(AlertOnContainerChange())
    set_mock_value(robot.current_puck, 5)

    mock_composite = MagicMock()
    mock_composite.robot = robot
    mock_composite.backlight = backlight

    patched_robot_load.side_effect = NotImplementedError()

    with pytest.raises(NotImplementedError):
        RE(
            wrap_plan(
                10,
                partial(
                    robot_load_and_snapshots,
                    mock_composite,
                    MagicMock(),
                    MagicMock(),
                    1,
                    2,
                    None,
                ),
            )
        )

    mock_alert_service.raise_alert.assert_called_once_with(
        "UDC moved on to puck 10 on i03",
        "Hyperion finished container 5 and moved on to 10",
        {
            Metadata.SAMPLE_ID: "10",
            Metadata.VISIT: "cm1234-67",
            Metadata.CONTAINER: "10",
        },
    )



from contextlib import nullcontext as does_not_raise
from pathlib import Path

import pytest
from pydantic import ValidationError

from mx_bluesky.common.parameters.components import WithSnapshot


@pytest.mark.parametrize(
    "model, expectation",
    [
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [],
                "use_grid_snapshots": False,
            },
            does_not_raise(),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [],
                "use_grid_snapshots": True,
            },
            does_not_raise(),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [10, 20, 30, 40],
                "use_grid_snapshots": False,
            },
            does_not_raise(),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [0, 270],
                "use_grid_snapshots": True,
            },
            pytest.raises(ValidationError),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [0],
                "use_grid_snapshots": True,
            },
            pytest.raises(ValidationError),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "snapshot_omegas_deg": [10, 80],
                "use_grid_snapshots": True,
            },
            pytest.raises(ValidationError),
        ],
        [
            {
                "snapshot_directory": Path("/tmp"),
                "use_grid_snapshots": True,
            },
            does_not_raise(),
        ],
    ],
)
def test_validate_with_snapshot_omegas_grid_snapshots(model, expectation):
    with expectation:
        WithSnapshot.model_validate(model)



import json
import signal
import threading
from dataclasses import asdict
from sys import argv
from traceback import format_exception

from blueapi.core import BlueskyContext
from flask import Flask, request
from flask_restful import Api, Resource

from mx_bluesky.common.external_interaction import alerting
from mx_bluesky.common.external_interaction.alerting.log_based_service import (
    LoggingAlertService,
)
from mx_bluesky.common.parameters.constants import Actions, Status
from mx_bluesky.common.utils.log import (
    LOGGER,
    do_default_logging_setup,
    flush_debug_handler,
)
from mx_bluesky.hyperion.baton_handler import run_forever
from mx_bluesky.hyperion.experiment_plans.experiment_registry import (
    PLAN_REGISTRY,
    PlanNotFound,
)
from mx_bluesky.hyperion.external_interaction.agamemnon import (
    compare_params,
    update_params_from_agamemnon,
)
from mx_bluesky.hyperion.parameters.cli import (
    HyperionArgs,
    HyperionMode,
    parse_cli_args,
)
from mx_bluesky.hyperion.parameters.constants import CONST, HyperionConstants
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.plan_runner import PlanRunner
from mx_bluesky.hyperion.plan_runner_api import create_server_for_udc
from mx_bluesky.hyperion.runner import (
    GDARunner,
    StatusAndMessage,
    make_error_status_and_message,
)
from mx_bluesky.hyperion.utils.context import setup_context


def compose_start_args(context: BlueskyContext, plan_name: str, action: Actions):
    experiment_registry_entry = PLAN_REGISTRY.get(plan_name)
    if experiment_registry_entry is None:
        raise PlanNotFound(f"Experiment plan '{plan_name}' not found in registry.")

    experiment_internal_param_type = experiment_registry_entry.get("param_type")
    plan = context.plan_functions.get(plan_name)
    if experiment_internal_param_type is None:
        raise PlanNotFound(
            f"Corresponding internal param type for '{plan_name}' not found in registry."
        )
    if plan is None:
        raise PlanNotFound(
            f"Experiment plan '{plan_name}' not found in context. Context has {context.plan_functions.keys()}"
        )
    try:
        parameters = experiment_internal_param_type(**json.loads(request.data))
        parameters = update_params_from_agamemnon(parameters)
        if isinstance(parameters, LoadCentreCollect):
            compare_params(parameters)
        if parameters.model_extra:
            raise ValueError(f"Extra fields not allowed {parameters.model_extra}")
    except Exception as e:
        raise ValueError(
            f"Supplied parameters don't match the plan for this endpoint {request.data}, for plan {plan_name}"
        ) from e
    return plan, parameters, plan_name


class RunExperiment(Resource):
    def __init__(self, runner: GDARunner, context: BlueskyContext) -> None:
        super().__init__()
        self.runner = runner
        self.context = context

    def put(self, plan_name: str, action: Actions):
        status_and_message = StatusAndMessage(Status.FAILED, f"{action} not understood")
        if action == Actions.START.value:
            try:
                plan, params, plan_name = compose_start_args(
                    self.context, plan_name, action
                )
                status_and_message = self.runner.start(plan, params, plan_name)
            except Exception as e:
                status_and_message = make_error_status_and_message(e)
                LOGGER.error("".join(format_exception(e)))

        elif action == Actions.STOP.value:
            status_and_message = self.runner.stop()
        # no idea why mypy gives an attribute error here but nowhere else for this
        # exact same situation...
        return asdict(status_and_message)  # type: ignore


class StopOrStatus(Resource):
    def __init__(self, runner: GDARunner) -> None:
        super().__init__()
        self.runner: GDARunner = runner

    def put(self, action):
        status_and_message = StatusAndMessage(Status.FAILED, f"{action} not understood")
        if action == Actions.STOP.value:
            status_and_message = self.runner.stop()
        return asdict(status_and_message)

    def get(self, **kwargs):
        action = kwargs.get("action")
        status_and_message = StatusAndMessage(Status.FAILED, f"{action} not understood")
        if action == Actions.STATUS.value:
            LOGGER.debug(
                f"Runner received status request - state of the runner object is: {self.runner.__dict__} - state of the RE is: {self.runner.RE.__dict__}"
            )
            status_and_message = self.runner.current_status
        return asdict(status_and_message)


class FlushLogs(Resource):
    def put(self, **kwargs):
        try:
            log_file = flush_debug_handler()
            status_and_message = StatusAndMessage(
                Status.SUCCESS, f"Flushed debug log to {log_file}"
            )
        except Exception as e:
            status_and_message = StatusAndMessage(
                Status.FAILED, f"Failed to flush debug log: {e}"
            )
        return asdict(status_and_message)


def create_app(runner: GDARunner, test_config=None) -> Flask:
    app = Flask(__name__)
    if test_config:
        app.config.update(test_config)
    api = Api(app)

    api.add_resource(
        RunExperiment,
        "/<string:plan_name>/<string:action>",
        resource_class_args=[runner, runner.context],
    )

    api.add_resource(
        FlushLogs,
        "/flush_debug_log",
    )
    api.add_resource(
        StopOrStatus,
        "/<string:action>",
        resource_class_args=[runner],
    )
    return app


def initialise_globals(args: HyperionArgs):
    """Do all early main low-level application initialisation."""
    do_default_logging_setup(
        CONST.LOG_FILE_NAME, CONST.GRAYLOG_PORT, dev_mode=args.dev_mode
    )
    LOGGER.info(f"Hyperion launched with args:{argv}")
    alerting.set_alerting_service(LoggingAlertService(CONST.GRAYLOG_STREAM_ID))


def main():
    """Main application entry point."""
    args = parse_cli_args()
    initialise_globals(args)
    hyperion_port = HyperionConstants.HYPERION_PORT
    context = setup_context(dev_mode=args.dev_mode)

    if args.mode == HyperionMode.GDA:
        runner = GDARunner(context=context)
        app = create_app(runner)
        flask_thread = threading.Thread(
            target=lambda: app.run(
                host="0.0.0.0", port=hyperion_port, debug=True, use_reloader=False
            ),
            daemon=True,
        )
        flask_thread.start()
        LOGGER.info(
            f"Hyperion now listening on {hyperion_port} ({'IN DEV' if args.dev_mode else ''})"
        )
        runner.wait_on_queue()
    else:
        plan_runner = PlanRunner(context, args.dev_mode)
        create_server_for_udc(plan_runner)
        _register_sigterm_handler(plan_runner)
        run_forever(plan_runner)


def _register_sigterm_handler(runner: PlanRunner):
    def shutdown_on_sigterm(sig_num, frame):
        LOGGER.info("Received SIGTERM, shutting down...")
        runner.shutdown()

    signal.signal(signal.SIGTERM, shutdown_on_sigterm)


if __name__ == "__main__":
    main()



import bluesky.plan_stubs as bps
from bluesky.utils import MsgGenerator
from dodal.common import inject
from dodal.devices.aithre_lasershaping.goniometer import Goniometer
from dodal.devices.aithre_lasershaping.laser_robot import ForceBit, LaserRobot


def set_beamline_safe_on_robot(
    robot: LaserRobot = inject("robot"), goniometer: Goniometer = inject("goniometer")
) -> MsgGenerator:
    """
    The beamline safe PV is used in the Aithre laser shaping system to indicate whether
    the goniometer is in the correct position for the robot to load a sample. The robot
    is trained to load at the goniometer zero position, so if the translation and
    rotation axes of the goniometer are at zero, then the beamline safe PV bit is forced
    on.
    """
    pvs = [
        goniometer.x,
        goniometer.y,
        goniometer.z,
        goniometer.sampy,
        goniometer.sampz,
        goniometer.omega,
    ]

    values: list[float] = []
    for pv in pvs:
        values.append((yield from bps.rd(pv)))

    set_value = (
        ForceBit.ON.value
        if all(round(value, 3) == 0 for value in values)
        else ForceBit.NO.value
    )
    yield from bps.abs_set(robot.set_beamline_safe, set_value, wait=True)


def go_to_zero(
    goniometer: Goniometer = inject("goniometer"), group="move_to_zero", wait=True
) -> MsgGenerator:
    """
    Rotate the goniometer and set stages to zero in preparation for robot load/unload.
    Pass wait=False to avoid waiting for the move to complete.
    """
    yield from bps.abs_set(goniometer.omega, 0, group=group)
    yield from bps.abs_set(goniometer.x, 0, group=group)
    yield from bps.abs_set(goniometer.y, 0, group=group)
    yield from bps.abs_set(goniometer.z, 0, group=group)
    yield from bps.abs_set(goniometer.sampy, 0, group=group)
    yield from bps.abs_set(goniometer.sampz, 0, group=group)
    if wait:
        yield from bps.wait(group=group)



import os

from dodal.devices.detector import EIGER2_X_16M_SIZE
from pydantic.dataclasses import dataclass

from mx_bluesky.common.parameters.constants import (
    DeviceSettingsConstants,
    DocDescriptorNames,
    EnvironmentConstants,
    ExperimentParamConstants,
    FeatureSetting,
    FeatureSettingources,
    HardwareConstants,
    OavConstants,
    PlanGroupCheckpointConstants,
    PlanNameConstants,
)

TEST_MODE = os.environ.get("HYPERION_TEST_MODE")


@dataclass(frozen=True)
class I03Constants:
    BEAMLINE = "BL03S" if TEST_MODE else "BL03I"
    DETECTOR = EIGER2_X_16M_SIZE
    INSERTION_PREFIX = "SR03S" if TEST_MODE else "SR03I"
    OAV_CENTRING_FILE = OavConstants.OAV_CONFIG_JSON
    SHUTTER_TIME_S = 0.06
    USE_GPU_RESULTS = True
    OMEGA_FLIP = True
    ALTERNATE_ROTATION_DIRECTION = True


# These currently exist in GDA domain.properties
class HyperionFeatureSettingources(FeatureSettingources):
    USE_GPU_RESULTS = "gda.mx.hyperion.xrc.use_gpu_results"
    USE_PANDA_FOR_GRIDSCAN = "gda.mx.hyperion.use_panda_for_gridscans"
    SET_STUB_OFFSETS = "gda.mx.hyperion.do_stub_offsets"
    PANDA_RUNUP_DISTANCE_MM = "gda.mx.hyperion.panda_runup_distance_mm"


# Use these defaults if we can't read from the config server
@dataclass
class HyperionFeatureSetting(FeatureSetting):
    USE_GPU_RESULTS: bool = True
    USE_PANDA_FOR_GRIDSCAN: bool = False
    SET_STUB_OFFSETS: bool = False
    PANDA_RUNUP_DISTANCE_MM: float = 0.16


@dataclass(frozen=True)
class HyperionConstants:
    ZOCALO_ENV = EnvironmentConstants.ZOCALO_ENV
    HARDWARE = HardwareConstants()
    I03 = I03Constants()
    PARAM = ExperimentParamConstants()
    PLAN = PlanNameConstants()
    WAIT = PlanGroupCheckpointConstants()
    HYPERION_PORT = 5005
    CALLBACK_0MQ_PROXY_PORTS = (5577, 5578)
    DESCRIPTORS = DocDescriptorNames()
    CONFIG_SERVER_URL = (
        "http://fake-url-not-real"
        if TEST_MODE
        else "https://daq-config.diamond.ac.uk/api"
    )
    GRAYLOG_PORT = 12232  # Hyperion stream
    GRAYLOG_STREAM_ID = "66264f5519ccca6d1c9e4e03"
    PARAMETER_SCHEMA_DIRECTORY = "src/hyperion/parameters/schemas/"
    LOG_FILE_NAME = "hyperion.log"
    DEVICE_SETTINGS_CONSTANTS = DeviceSettingsConstants()


CONST = HyperionConstants()



from __future__ import annotations

from collections.abc import Generator

import bluesky.plan_stubs as bps
import numpy as np
import pydantic
from blueapi.core import BlueskyContext
from bluesky.preprocessors import run_decorator, set_run_key_decorator, subs_wrapper
from bluesky.utils import MsgGenerator
from dodal.devices.baton import Baton
from dodal.devices.oav.oav_parameters import OAVParameters

import mx_bluesky.common.xrc_result as flyscan_result
from mx_bluesky.common.parameters.components import WithSnapshot
from mx_bluesky.common.utils.context import device_composite_from_context
from mx_bluesky.common.utils.exceptions import CrystalNotFoundException
from mx_bluesky.common.utils.log import LOGGER
from mx_bluesky.common.xrc_result import XRayCentreEventHandler
from mx_bluesky.hyperion.experiment_plans.robot_load_then_centre_plan import (
    RobotLoadThenCentreComposite,
    robot_load_then_xray_centre,
)
from mx_bluesky.hyperion.experiment_plans.rotation_scan_plan import (
    RotationScan,
    RotationScanComposite,
    rotation_scan_internal,
)
from mx_bluesky.hyperion.external_interaction.config_server import (
    get_hyperion_config_client,
)
from mx_bluesky.hyperion.parameters.constants import CONST, I03Constants
from mx_bluesky.hyperion.parameters.load_centre_collect import LoadCentreCollect
from mx_bluesky.hyperion.parameters.rotation import RotationScanPerSweep


@pydantic.dataclasses.dataclass(config={"arbitrary_types_allowed": True})
class LoadCentreCollectComposite(RobotLoadThenCentreComposite, RotationScanComposite):
    """Composite that provides access to the required devices."""

    baton: Baton


def create_devices(context: BlueskyContext) -> LoadCentreCollectComposite:
    """Create the necessary devices for the plan."""
    return device_composite_from_context(context, LoadCentreCollectComposite)


def load_centre_collect_full(
    composite: LoadCentreCollectComposite,
    parameters: LoadCentreCollect,
    oav_params: OAVParameters | None = None,
) -> MsgGenerator:
    """Attempt a complete data collection experiment, consisting of the following:
    * Load the sample if necessary
    * Move to the specified goniometer start angles
    * Perform optical centring, then X-ray centring
    * If X-ray centring finds one or more diffracting centres then for each centre
     that satisfies the chosen selection function,
     move to that centre and do a collection with the specified parameters.
    """

    get_hyperion_config_client().refresh_cache()

    if not oav_params:
        oav_params = OAVParameters(context="xrayCentring")
    oav_config_file = oav_params.oav_config_json

    @set_run_key_decorator(CONST.PLAN.LOAD_CENTRE_COLLECT)
    @run_decorator(
        md={
            "metadata": {
                "sample_id": parameters.sample_id,
                "visit": parameters.visit,
                "container": parameters.sample_puck,
            },
            "activate_callbacks": [
                "BeamDrawingCallback",
                "SampleHandlingCallback",
                "AlertOnContainerChange",
            ],
            "with_snapshot": parameters.multi_rotation_scan.model_dump_json(
                include=WithSnapshot.model_fields.keys()  # type: ignore
            ),
        }
    )
    def plan_with_callback_subs():
        flyscan_event_handler = XRayCentreEventHandler()
        try:
            yield from subs_wrapper(
                robot_load_then_xray_centre(
                    composite, parameters.robot_load_then_centre, oav_config_file
                ),
                flyscan_event_handler,
            )
        except CrystalNotFoundException:
            if parameters.select_centres.ignore_xtal_not_found:
                LOGGER.info("Ignoring crystal not found due to parameter settings.")
            else:
                raise

        locations_to_collect_um: list[np.ndarray]
        samples_to_collect: list[int]

        if flyscan_event_handler.xray_centre_results:
            selection_func = flyscan_result.resolve_selection_fn(
                parameters.selection_params
            )
            hits = selection_func(flyscan_event_handler.xray_centre_results)
            hits_to_collect = []
            for hit in hits:
                if hit.sample_id is None:
                    LOGGER.warning(
                        f"Diffracting centre {hit} not collected because no sample id was assigned."
                    )
                else:
                    hits_to_collect.append(hit)

            locations_to_collect_um = [
                hit.centre_of_mass_mm * 1000 for hit in hits_to_collect
            ]
            samples_to_collect = [hit.sample_id for hit in hits_to_collect]
            LOGGER.info(
                f"Selected hits {hits_to_collect} using {selection_func}, args={parameters.selection_params}"
            )
        else:
            # If the xray centring hasn't found a result but has not thrown an error it
            # means that we do not need to recentre and can collect where we are
            initial_x_mm = yield from bps.rd(composite.smargon.x.user_readback)
            initial_y_mm = yield from bps.rd(composite.smargon.y.user_readback)
            initial_z_mm = yield from bps.rd(composite.smargon.z.user_readback)

            locations_to_collect_um = [
                np.array([initial_x_mm, initial_y_mm, initial_z_mm]) * 1000
            ]
            samples_to_collect = [parameters.sample_id]

        multi_rotation = parameters.multi_rotation_scan
        rotation_template = multi_rotation.rotation_scans.copy()

        multi_rotation.rotation_scans.clear()

        is_alternating = I03Constants.ALTERNATE_ROTATION_DIRECTION

        generator = rotation_scan_generator(is_alternating)
        next(generator)
        for location, sample_id in zip(
            locations_to_collect_um, samples_to_collect, strict=True
        ):
            for rot in rotation_template:
                combination = generator.send((rot, location, sample_id))
                multi_rotation.rotation_scans.append(combination)
        multi_rotation = RotationScan.model_validate(multi_rotation)

        assert (
            multi_rotation.demand_energy_ev
            == parameters.robot_load_then_centre.demand_energy_ev
        ), "Setting a different energy for gridscan and rotation is not supported"
        yield from rotation_scan_internal(composite, multi_rotation, oav_params)

    yield from plan_with_callback_subs()


def rotation_scan_generator(
    is_alternating: bool,
) -> Generator[
    RotationScanPerSweep, tuple[RotationScanPerSweep, np.ndarray, int], None
]:
    scan_template, location, sample_id = yield  # type: ignore
    next_rotation_direction = scan_template.rotation_direction
    while True:
        scan = scan_template.model_copy()
        (
            scan.x_start_um,
            scan.y_start_um,
            scan.z_start_um,
        ) = location
        scan.sample_id = sample_id
        if is_alternating:
            if next_rotation_direction != scan.rotation_direction:
                # If originally specified direction of the current scan is different
                # from that required, swap the start and ends.
                start = scan.omega_start_deg
                rotation_sign = scan.rotation_direction.multiplier
                end = start + rotation_sign * scan.scan_width_deg
                scan.omega_start_deg = end
                scan.rotation_direction = next_rotation_direction
            next_rotation_direction = next_rotation_direction.opposite

        scan_template, location, sample_id = yield scan



import pytest
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from dodal.devices.aperturescatterguard import ApertureScatterguard, ApertureValue
from dodal.devices.smargon import CombinedMove, Smargon
from ophyd_async.testing import get_mock_put

from mx_bluesky.common.device_setup_plans.manipulate_sample import (
    move_aperture_if_required,
    move_phi_chi_omega,
    move_x_y_z,
)


@pytest.mark.parametrize(
    "set_position",
    [
        (ApertureValue.SMALL),
        (ApertureValue.MEDIUM),
        (ApertureValue.OUT_OF_BEAM),
        (ApertureValue.LARGE),
    ],
)
async def test_move_aperture_goes_to_correct_position(
    aperture_scatterguard: ApertureScatterguard,
    RE: RunEngine,
    set_position,
):
    RE(move_aperture_if_required(aperture_scatterguard, set_position))
    last_pos = get_mock_put(aperture_scatterguard.selected_aperture).call_args[0]
    assert last_pos == (set_position,)


async def test_move_aperture_does_nothing_when_none_selected(
    aperture_scatterguard: ApertureScatterguard, RE: RunEngine
):
    get_mock_put(aperture_scatterguard.selected_aperture).reset_mock()
    RE(move_aperture_if_required(aperture_scatterguard, None))
    mock_put = get_mock_put(aperture_scatterguard.selected_aperture)
    mock_put.assert_not_called()


def test_move_x_y_z_no_wait(
    smargon: Smargon,
    sim_run_engine: RunEngineSimulator,
):
    msgs = sim_run_engine.simulate_plan(move_x_y_z(smargon, 10.0, 5.0, None))
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == smargon.name
        and msg.args[0] == CombinedMove(x=10.0, y=5.0, z=None),
    )
    assert len(msgs) == 1


def test_move_x_y_z_wait(
    smargon: Smargon,
    sim_run_engine: RunEngineSimulator,
):
    msgs = sim_run_engine.simulate_plan(move_x_y_z(smargon, 10.0, 5.0, None, wait=True))
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == smargon.name
        and msg.args[0] == CombinedMove(x=10.0, y=5.0, z=None),
    )
    group = msgs[0].kwargs["group"]
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == group,
    )


def test_move_phi_chi_omega_no_wait(
    smargon: Smargon,
    sim_run_engine: RunEngineSimulator,
):
    msgs = sim_run_engine.simulate_plan(move_phi_chi_omega(smargon, 10.0, 5.0, None))
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == smargon.name
        and msg.args[0] == CombinedMove(phi=10.0, chi=5.0, omega=None),
    )
    assert len(msgs) == 1


def test_move_phi_chi_omega_wait(
    smargon: Smargon,
    sim_run_engine: RunEngineSimulator,
):
    msgs = sim_run_engine.simulate_plan(
        move_phi_chi_omega(smargon, 10.0, 5.0, None, wait=True)
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == smargon.name
        and msg.args[0] == CombinedMove(phi=10.0, chi=5.0, omega=None),
    )
    group = msgs[0].kwargs["group"]
    assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "wait" and msg.kwargs["group"] == group,
    )



from functools import partial
from unittest.mock import ANY, MagicMock, call, patch

import pytest
from bluesky import Msg
from bluesky.run_engine import RunEngine
from bluesky.simulators import RunEngineSimulator, assert_message_and_return_remaining
from bluesky.utils import MsgGenerator
from dodal.devices.aperturescatterguard import ApertureScatterguard
from dodal.devices.attenuator.attenuator import BinaryFilterAttenuator
from dodal.devices.backlight import Backlight
from dodal.devices.common_dcm import BaseDCM
from dodal.devices.detector.detector_motion import DetectorMotion
from dodal.devices.eiger import EigerDetector
from dodal.devices.fast_grid_scan import (
    ZebraFastGridScan,
)
from dodal.devices.flux import Flux
from dodal.devices.mx_phase1.beamstop import Beamstop
from dodal.devices.oav.oav_detector import OAV
from dodal.devices.oav.pin_image_recognition import PinTipDetection
from dodal.devices.robot import BartRobot
from dodal.devices.s4_slit_gaps import S4SlitGaps
from dodal.devices.smargon import Smargon
from dodal.devices.synchrotron import Synchrotron
from dodal.devices.undulator import Undulator
from dodal.devices.xbpm_feedback import XBPMFeedback
from dodal.devices.zebra.zebra import Zebra
from dodal.devices.zebra.zebra_controlled_shutter import ZebraShutter
from dodal.devices.zocalo import ZocaloResults
from tests.conftest import TEST_RESULT_LARGE, simulate_xrc_result
from tests.unit_tests.common.experiment_plans.test_common_flyscan_xray_centre_plan import (
    CompleteException,
)

from mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan import (
    get_ready_for_oav_and_close_shutter,
    i04_grid_detect_then_xray_centre,
)
from mx_bluesky.common.parameters.constants import PlanNameConstants
from mx_bluesky.common.parameters.gridscan import GridCommon


class CustomException(Exception): ...


@pytest.fixture
def i04_grid_detect_then_xrc_default_params(
    aperture_scatterguard: ApertureScatterguard,
    attenuator: BinaryFilterAttenuator,
    backlight: Backlight,
    beamstop_phase1: Beamstop,
    dcm: BaseDCM,
    zebra_fast_grid_scan: ZebraFastGridScan,
    flux: Flux,
    oav: OAV,
    pin_tip_detection_with_found_pin: PinTipDetection,
    s4_slit_gaps: S4SlitGaps,
    undulator: Undulator,
    xbpm_feedback: XBPMFeedback,
    zebra: Zebra,
    robot: BartRobot,
    sample_shutter: ZebraShutter,
    eiger: EigerDetector,
    synchrotron: Synchrotron,
    zocalo: ZocaloResults,
    smargon: Smargon,
    detector_motion: DetectorMotion,
    test_full_grid_scan_params: GridCommon,
):
    return partial(
        i04_grid_detect_then_xray_centre,
        parameters=test_full_grid_scan_params,
        aperture_scatterguard=aperture_scatterguard,
        attenuator=attenuator,
        backlight=backlight,
        beamstop=beamstop_phase1,
        dcm=dcm,
        zebra_fast_grid_scan=zebra_fast_grid_scan,
        flux=flux,
        oav=oav,
        pin_tip_detection=pin_tip_detection_with_found_pin,
        s4_slit_gaps=s4_slit_gaps,
        undulator=undulator,
        xbpm_feedback=xbpm_feedback,
        zebra=zebra,
        robot=robot,
        sample_shutter=sample_shutter,
        eiger=eiger,
        synchrotron=synchrotron,
        zocalo=zocalo,
        smargon=smargon,
        detector_motion=detector_motion,
    )


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.setup_beamline_for_OAV",
    autospec=True,
)
def test_get_ready_for_oav_and_close_shutter_closes_shutter_and_calls_setup_for_oav_plan(
    mock_setup_beamline_for_oav: MagicMock,
    sim_run_engine: RunEngineSimulator,
    grid_detect_xrc_devices,
):
    mock_setup_beamline_for_oav.return_value = iter([Msg("setup_beamline_for_oav")])

    msgs = sim_run_engine.simulate_plan(
        get_ready_for_oav_and_close_shutter(
            grid_detect_xrc_devices.smargon,
            grid_detect_xrc_devices.backlight,
            grid_detect_xrc_devices.aperture_scatterguard,
            grid_detect_xrc_devices.detector_motion,
        )
    )
    msgs = assert_message_and_return_remaining(
        msgs, predicate=lambda msg: msg.command == "wait"
    )

    msgs = assert_message_and_return_remaining(
        msgs, predicate=lambda msg: msg.command == "setup_beamline_for_oav"
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        predicate=lambda msg: msg.command == "set"
        and msg.obj.name == "detector_motion-shutter"
        and msg.args[0] == 0,
    )
    msgs = assert_message_and_return_remaining(
        msgs, predicate=lambda msg: msg.command == "wait"
    )


@pytest.mark.parametrize(
    "udc",
    [(True), (False)],
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_ready_for_oav_and_close_shutter",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.grid_detect_then_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.setup_beamline_for_OAV",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
def test_i04_grid_detect_then_xrc_closes_shutter_and_tidies_if_not_udc(
    mock_create_gridscan_callbacks: MagicMock,
    mock_setup_beamline_for_oav: MagicMock,
    mock_grid_detect_then_xray_centre: MagicMock,
    mock_get_ready_for_oav_and_close_shutter: MagicMock,
    RE: RunEngine,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
    udc: bool,
):
    RE(
        i04_grid_detect_then_xrc_default_params(
            udc=udc,
        )
    )

    call_count = 0 if udc else 1

    assert mock_get_ready_for_oav_and_close_shutter.call_count == call_count


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
@patch(
    "mx_bluesky.common.preprocessors.preprocessors.check_and_pause_feedback",
    autospec=True,
)
@patch(
    "mx_bluesky.common.preprocessors.preprocessors.unpause_xbpm_feedback_and_set_transmission_to_1",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.common_flyscan_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
    autospec=True,
)
@patch("bluesky.plan_stubs.sleep", autospec=True)
def test_i04_xray_centre_unpauses_xbpm_feedback_on_exception(
    mock_sleep: MagicMock,
    mock_grid_detection_callback: MagicMock,
    mock_create_parameters_for_flyscan_xray_centre: MagicMock,
    mock_grid_detection_plan: MagicMock,
    mock_common_flyscan_xray_centre: MagicMock,
    mock_unpause_and_set_transmission: MagicMock,
    mock_check_and_pause: MagicMock,
    mock_create_gridscan_callbacks: MagicMock,
    RE: RunEngine,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
):
    mock_common_flyscan_xray_centre.side_effect = CustomException

    with pytest.raises(CustomException):  # noqa: B017
        RE(i04_grid_detect_then_xrc_default_params())

    # Called once on exception and once on close_run
    mock_unpause_and_set_transmission.assert_has_calls([call(ANY, ANY)])


@patch("bluesky.plan_stubs.sleep", autospec=True)
@patch(
    "mx_bluesky.common.experiment_plans.inner_plans.do_fgs.check_topup_and_wait_if_necessary",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.XRayCentreEventHandler"
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.change_aperture_then_move_to_xtal"
)
def test_i04_grid_detect_then_xray_centre_pauses_and_unpauses_xbpm_feedback_in_correct_order(
    mock_change_aperture_then_move: MagicMock,
    mock_events_handler: MagicMock,
    mock_create_parameters: MagicMock,
    mock_grid_detection_callback: MagicMock,
    mock_grid_detection_plan: MagicMock,
    mock_check_topup: MagicMock,
    mock_wait: MagicMock,
    sim_run_engine: RunEngineSimulator,
    zocalo: ZocaloResults,
    hyperion_fgs_params,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
):
    flyscan_event_handler = MagicMock()
    flyscan_event_handler.xray_centre_results = "dummy"
    mock_events_handler.return_value = flyscan_event_handler
    mock_create_parameters.return_value = hyperion_fgs_params
    simulate_xrc_result(
        sim_run_engine,
        zocalo,
        TEST_RESULT_LARGE,
    )

    msgs = sim_run_engine.simulate_plan(
        i04_grid_detect_then_xrc_default_params(),
    )

    # Assert order: pause -> open run -> close run -> unpause (set attenuator)
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "trigger" and msg.obj.name == "xbpm_feedback",
    )
    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "open_run"
        and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "close_run"
        and msg.run == PlanNameConstants.GRIDSCAN_OUTER,
    )

    msgs = assert_message_and_return_remaining(
        msgs,
        lambda msg: msg.command == "set"
        and msg.obj.name == "attenuator"
        and msg.args == (1.0,),
    )


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.grid_detection_plan",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.move_aperture_if_required",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.GridDetectionCallback",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_grid_detect_then_xray_centre_plan.create_parameters_for_flyscan_xray_centre",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan.run_gridscan",
)
@patch(
    "mx_bluesky.common.experiment_plans.common_flyscan_xray_centre_plan._fetch_xrc_results_from_zocalo",
)
@patch(
    "dodal.plans.preprocessors.verify_undulator_gap.verify_undulator_gap",
)
def test_i04_grid_detect_then_xray_centre_does_undulator_check_before_collection(
    mock_verify_gap: MagicMock,
    mock_fetch_zocalo_results: MagicMock,
    mock_run_gridscan: MagicMock,
    mock_create_parameters: MagicMock,
    mock_grid_params_callback: MagicMock,
    mock_move_aperture_if_required: MagicMock,
    mock_grid_detection_plan: MagicMock,
    mock_create_gridscan_callbacks: MagicMock,
    RE: RunEngine,
    hyperion_fgs_params,
    i04_grid_detect_then_xrc_default_params: partial[MsgGenerator],
):
    mock_create_parameters.return_value = hyperion_fgs_params
    mock_run_gridscan.side_effect = CompleteException
    with pytest.raises(CompleteException):
        RE(i04_grid_detect_then_xrc_default_params())

    mock_verify_gap.assert_called_once()


@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.get_ready_for_oav_and_close_shutter",
    autospec=True,
)
@patch(
    "mx_bluesky.beamlines.i04.experiment_plans.i04_grid_detect_then_xray_centre_plan.create_gridscan_callbacks",
    autospec=True,
)
def test_i04_grid_detect_then_xrc_tidies_up_on_exception(
    mock_create_gridscan_callbacks: MagicMock,
    mock_get_ready_for_oav_and_close_shutter: MagicMock,
    RE: RunEngine,
    i04_grid_detect_then_xrc_default_params,
):
    mock_create_gridscan_callbacks.side_effect = CustomException
    with pytest.raises(CustomException):
        RE(
            i04_grid_detect_then_xrc_default_params(
                udc=False,
            )
        )

    assert mock_get_ready_for_oav_and_close_shutter.call_count == 1
